#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

class Pivm < ActiveRecord::Migration
  
  def self.up
  
    create_table 'pivm.vm_queue' do |t|
		t.column :process_id,	:integer,	:null => false
		t.column :operation,	:text,		:null => false
		t.column :channel,		:text
		t.column :thread_id,	:integer
		t.column :is_replication,	:boolean,	:default => false
	end
	
	add_index "pivm.vm_queue", [:operation, :channel]

    execute %q{
		create or replace function pivm.vm_interact(IN uri text, IN verb text, IN request text, IN thread_id int4, OUT process_id int4, OUT output_channel text, OUT response xenodot.node) as $ruby$
			init_xenodot
			
			# v o
			output_channel = PiVM.fresh_name("o")
			
			# i<output_channel, verb, request>
			process_id = PiVM.find_process(:channel => uri, :operation => 'read')
			if process_id
				# do communicate
				PiVM.communicate(process_id, nil, :output_channel => output_channel, :verb=> verb, :request => request)
			else
				# TODO create process node with i<output_channel, verb, request>
				
				# enque process
			end
			
		
		$ruby$ language plruby;
	}

 	execute %q{create or replace function pivm.vm_init() returns void as $ruby$
 			init_xenodot
 			$DB = true
 			require 'pivm'
 			
 			$ruby$ language plruby
 			}
 	
    execute %q{
		create or replace function pivm.vm_next_step(INOUT process_id int4, IN thread_id int4, IN output_channel text,  OUT response text) as $ruby$
			
			response = nil
			#check if the response is ready: o(response)

			p = PiVM::Queue.dequeue('write',output_channel)
			if p
				warn "found result"
				response = PiVM::Pi.write(p)
				PiVM::Queue.enqueue(p)
			else
				#search for processes to handle
				if process_id

					p = Xenodot::Node.new(process_id)
					p.act_as PiVM::Process
				else

					p = PiVM::Queue.dequeue
					if p
						process_id = p.id
					else
						sleep 1
						return 
					end
				end
				#handle one prefix of the process
				warn "process is nil" if p.nil?
				unless PiVM.handle_process(p)
					process_id = nil
				end
			end
			
			return process_id, response
		$ruby$ language plruby;
	}
	
	# Retrieve a process from the waiting queue that match the query
	# This function should support maximal parallelism, that mean that no process
	# should wait for locks
	# Replicative processes must not be removed from the queue and should not be locked at
	# any time
	execute %q{
	create or replace function pivm.vm_dequeue(operation text, channel text, replication boolean) returns SETOF xenodot.node as $pgsql$
		DECLARE
			proc pivm.vm_queue%ROWTYPE;
			node xenodot.node;
			id int4;
			off int4 default 0;
		BEGIN
			LOOP
				IF replication THEN
					select * into proc from pivm.vm_queue as q where q.operation = operation and q.channel = channel limit 1 offset off;
				ELSE
					select * into proc from pivm.vm_queue as q where q.operation = operation and q.channel = channel and q.is_replication = false limit 1 offset off;
				END IF;
				IF NOT FOUND THEN
					RETURN;
				END IF;
				id := proc.process_id;
				IF proc.is_replication THEN
					EXIT;
				ELSE
					delete from pivm.vm_queue as q where q.process_id = id;
					IF FOUND THEN
						EXIT;
					ELSE
						RAISE WARNING 'Sombody deleted the rows before me the % time', off;
						off := off + 1;
						IF off > 100 THEN
							RAISE EXCEPTION 'Could not get and delete a row. Give up!';
						END IF;
					END IF;
				END IF;
			END LOOP;
			
			select * INTO node from xenodot.node as n where n.id = id;
			RETURN NEXT node;
			RETURN;
			
		END;
	$pgsql$ language plpgsql;
	}
	
	execute %q{ create or replace function pivm.profile_start() returns void as $ruby$
			require 'profiler'
			Profiler__::start_profile
	$ruby$ language plruby
	}
	
	execute %q{ create or replace function pivm.profile_stop(filename text) returns void as $ruby$
			Profiler__::stop_profile
			f = File.new(filename, "w+")
			f.chmod(0666)
			Profiler__::print_profile(f)
			f.close
	$ruby$ language plruby
	}
	
	begin
		execute "create sequence pivm.vm_id start 1000"
	rescue
		puts $!
	end
  end
  
  def self.down
  	drop_table 'pivm.vm_queue'
  end

  
end


