class Bootstrap < Xenodot::Migration
  
  def self.up
  
    # PL/SQL
    execute "create language plpgsql;" rescue {}
    
    # PL/Ruby
    # Assumes the script to run local on your db, change if necessary (full path to plruby.{so,bundle,dll,sl})
    plruby_lib = Dir.glob("{#{($LOAD_PATH).join(',')}}/plruby.{so,bundle,dll,sl}")[0]
    begin
      execute <<-SQL
        create function plruby_call_handler() returns language_handler
        as '#{plruby_lib}'
        language 'C';
        
        create trusted language 'plruby' handler plruby_call_handler lancompiler 'PL/Ruby';
      SQL
    rescue => ex
      puts "ERROR:"
      puts ex
      puts "Hint: Either you don't have pl/ruby or your are installing remotely, you will need to edit db/migration/001_bootstrap.rb"
    end
    
    create_table 'public.plruby_singleton_methods' do |t|
      t.column :name, :text, :null => false
      t.column :args, :text, :null => false
      t.column :body, :text, :null => false
    end

    # Helper functions
    execute <<-SQL
      create or replace function xenodot.has_value(anyelement) returns integer as $sql$
        select case
          when $1 is null then 0 
          else 1 
        end;
      $sql$ language sql immutable;
      
      create or replace function xenodot.is_true(boolean) returns integer as $sql$
        select case
          when $1 then 1
          else 0
        end;
      $sql$ language sql immutable;
    SQL

    create_singleton :init_xenodot do
      sql = %q{end
      
      $KCODE = 'u'
      require 'jcode'
      require 'open-uri'
      require 'set'
      require 'iconv'
      require 'htree'
      require 'rexml/functions'
      require 'rexml/xpath_parser'
      require 'set'
      }

      %w{
        dom/data_types
        query/sqlbuilder 
        query/xpath
        query/jcr_path 
        backend/hierarchy 
        backend/xml_parser 
        server/uri_templating
        dom/data_types
        dom/query
        dom/node
        
      }.each do |name|
        name += '.rb'
        sql += %Q{
          
          ### #{name}
          
        }
        
        sql += File.read(File.join(File.dirname(__FILE__), '../../lib/xenodot', name))
      end
      
      sql += %q{
      
      module Xenodot
        module Backend
          def self.sql(statement, values, types, options={})
            debug = options.key?(:debug) ? options[:debug] : false
            if debug
              warn ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"
              warn "Query:   #{statement}"
              warn "Params:  #{values.inspect}"
              warn "Types:   #{types.inspect}"
            end
            options['output'] = (options['output'] || options[:output] || :value).to_s
            options.delete(:output)
            options['count']  = options[:count] if options.key? :count
            options.delete(:count)
            warn "Options: #{options.inspect}" if debug
            
            query = $Plans[statement.to_sym] ||= PL::Plan.new(statement, types.collect{|type| type.to_s}, options).save
            if options[:cursor]
              begin
                return query.cursor('ruby', values)
              ensure
                if debug
                  warn "Returning cursor!"
                  warn "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"
                end
              end
            end

            result = query.exec(values, options)
            if debug
              warn "Result: #{result.inspect}"
              warn "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"
            end
            return result
          end
          
          def self.transaction! &block
            self.subtransaction! &block
          end
          
          def self.subtransaction!
            begin
              break_called = true
              yield
              break_called = false
            ensure
              raise 'Postgres does not currently support subtransactions via SAVEPOINT, hence no "break"' if break_called
            end
          end
          
          def self.transaction?
            true
          end
        end
      }
    end
    
  end
  
  def self.down
    # Taken care of by rake cleandb
  end

end