class Names < Xenodot::Migration
  def self.up
    create_table 'xenodot.uri' do |t|
      t.column  :prefix,      :text,  :null => false
      t.column  :value,       :text,  :null => false
    end
    
    execute <<-SQL
      alter table xenodot.uri alter column prefix set storage main;
      alter table xenodot.uri alter column value set storage main;
      create unique index uri_prefix on xenodot.uri(prefix);
      create unique index uri_value on xenodot.uri(value);
      cluster uri_value on xenodot.uri;
    SQL
    
    execute <<-SQL
      create or replace function xenodot.default_uri_prefix() returns text as $sql$
        select 'ns' || currval('xenodot.uri_id_seq');
      $sql$ language sql;
      
      alter table xenodot.uri alter column prefix set default xenodot.default_uri_prefix();
    SQL
    
    execute <<-SQL
      insert into xenodot.uri(id, prefix, value)  values(0,  '',       '');
      insert into xenodot.uri(id, prefix, value)  values(-1, 'xen',    '!xenodot!');

      insert into xenodot.uri(prefix, value)      values('xhtml',  'http://www.w3.org/1999/xhtml');
    SQL

    execute %q{
      create or replace function xenodot.uri_hierarchy(uri text) returns text as $ruby$
        require 'uri'

        uri = URI.parse(uri)

        hier = ''

        if uri.host
          host_tokens = uri.host.split('.').reverse
          start = 1
          start = 2 if uri.host =~ /\.(de\.com|eu\.com|or\.at|co\.at|com\.es|org\.es|biz\.pl|com\.pl|info\.pl|net\.pl|org\.pl|com\.pt|com\.pt|com\.ro|co\.uk|me\.uk|org\.uk|gb\.com|com\.ag|net\.ag|org\.ag|com\.mx|us\.com|com\.br|com\.uy|cn\.com|com\.cn|net\.cn|org\.cn|co\.il|co\.in|net\.in|org\.in|firm\.in|gen\.in|ind\.in|com\.sg|com\.tw|idv\.tw|org\.tw|com\.sc|net\.sc|org\.sc|co\.za|web\.com)\z/

          if host_tokens.size > 1
            top_level = host_tokens[0...start].reverse.join('.')
            hier << "#t:#{top_level}" if top_level

            domain = host_tokens[start]
            hier << "#d:#{domain}" if domain

            host_tokens[(start+1)..-1].reverse.each do |hostname|
              hier << "#h:#{hostname}"
            end
          else
            hier << "#h:#{uri.host}"
          end
        end

        # scheme is irrelevant to the hierarchy from here on.
        # hier << "#a:#{uri.scheme}" if uri.scheme
        
        hier << "#o:#{uri.port}" if uri.port

        if uri.path
          path_elements = uri.path.scan(/\/?(.+?)(\/|\Z)/)
          if path_elements.size > 0
            path_elements[0...-1].each do |pathkey|
              hier << "#p:#{pathkey}"
            end
            pathelement, suffix = path_elements[-1].to_s.scan(/(.*)\.(.+\Z)|(.+)/).flatten.compact
            hier << "#p:#{pathelement}"
            hier << "#x:#{suffix}" if suffix
          end  
        end

        if uri.query
          # normalize the query, taken from mongrel/camping
          params = {}
          uri.query.split(/[&;] */n).inject(params) { |h,p|
            k, v = p.tr('+', ' ').gsub(/((?:%[0-9a-fA-F]{2})+)/n){
              [$1.delete('%')].pack('H*')
            }.split('=',2)
            if cur = params[k]
              if cur.class == Array
                params[k] << v
              else
                params[k] = [cur, v]
              end
            else
              params[k] = v
            end
          }

          sortable_query = "#{params.keys.sort.collect{|key| "#{key.inspect} => #{params[key].inspect}"}.join(', ')}"
          hier << "#q:{#{URI.encode(sortable_query)}}"
        end

        hier << "#f:#{uri.fragment}" if uri.fragment

        return hier
      $ruby$ language plruby immutable;
    }
    
    execute %q{
      create or replace function xenodot.uri(uristring text) returns xenodot.uri as $pgsql$
        declare
          result xenodot.uri;
        begin
          select * into result from xenodot.uri where value = uristring or prefix = uristring limit 1;
          if not found then
            raise warning 'Uri[%] does not exist!', uristring;
          end if;
          return result;
        end;
      $pgsql$ language plpgsql stable;
      
      create or replace function xenodot.ensure_uri(text) returns xenodot.uri as $pgsql$
        declare
          parent xenodot.uri;
          result xenodot.uri;
        begin
          if $1 is null then
            select * into result from xenodot.uri where id = 0;
          else
            select * into result from xenodot.uri where prefix = $1 or value = $1 limit 1;
          end if;
          if not found then
            parent := xenodot.ensure_uri(substring($1 from '(.+)[:/#]+.*'));
            select * into parent from xenodot.uri where id = parent.id for share;
            select * into result from xenodot.uri where prefix = $1 or value = $1 limit 1;
            if not found then
              insert into xenodot.uri(value) values($1) returning * into result;
            end if;
          end if;
          return result;
        end;
      $pgsql$ language plpgsql;
    }

    create_table 'xenodot.qname' do |t|
      t.column  :ns,          :text,  :null => false
      t.column  :local_name,  :text,  :null => false
    end

    execute <<-SQL
      create unique index qname_value on xenodot.qname(ns, local_name);
      alter table xenodot.qname alter column ns         set storage main;
      alter table xenodot.qname alter column local_name set storage main;
      insert into xenodot.qname(id, ns, local_name) values( 0, '',  '');
      insert into xenodot.qname(id, ns, local_name) values(-1, '!xenodot!', '!document-root!');
    SQL
    
    execute <<-SQL
      create or replace function xenodot.ensure_qname(text, text) returns xenodot.qname as $pgsql$
        declare
          uri xenodot.uri;
          result xenodot.qname;
        begin
          if $1 is null or $2 is null then
            select * into result from xenodot.qname where id = 0;
          else
            select qname.* into result from xenodot.qname where ns = $1 and local_name = $2;
          end if;
          if not found then
            uri := xenodot.ensure_uri($1);
            select * into uri from xenodot.uri where id = uri.id for share;
            select qname.* into result from xenodot.qname where ns = uri.value and local_name = $2;

            if not found then
              insert into xenodot.qname(ns, local_name) values(uri.value, $2) returning * into result;
            end if;
          end if;
          return result;
        end;
      $pgsql$ language plpgsql;
    SQL
    
    execute <<-SQL
      create or replace function xenodot.qname(text, text) returns xenodot.qname as $pgsql$
        declare
          result xenodot.qname;
        begin
          select    qname.*
          into      result
          from      xenodot.qname, xenodot.uri
          where     (uri.prefix = $1 or uri.value = $1)
                and uri.value = qname.ns
                and qname.local_name = $2;
          
          if not found then
            raise warning 'QName[%, %] does not exist!', $1, $2;
          end if;
          
          return result;
        end;
      $pgsql$ language plpgsql stable;
    SQL
    
    execute <<-SQL
      create or replace function xenodot.prefix(xenodot.qname) returns text as $sql$
        select prefix from xenodot.uri where value = $1.ns;
      $sql$ language sql stable;
    SQL
  end
  
  def self.down
  end
end
