class Values < Xenodot::Migration
  def self.up

    $INLINED_TYPES.merge($REFERENCED_TYPES).each do |name, sql|
      execute <<-SQL
        create or replace function xenodot.type_#{name}() returns "char" as $sql$
          select '#{sql[0]}'::"char";
        $sql$ language sql immutable;
      SQL
    end
    
    valid_chars = $INLINED_TYPES.merge($REFERENCED_TYPES).collect{|name, sql| %Q{'#{sql[0]}'::"char"}}
    execute <<-SQL
      create or replace function xenodot.valid_type("char") returns boolean as $sql$
        select $1 in (#{valid_chars.join(',')})
      $sql$ language sql immutable;

      alter table xenodot.property_head add constraint known_value_type check((value).type is null or xenodot.valid_type((value).type));
    SQL
    
    $INLINED_TYPES.each do |name, sql|
      execute <<-SQL
        create or replace function xenodot.value_#{name}(#{sql[1].join(',')}) returns xenodot.value as $ruby$
          return '#{sql[0]}', #{sql[1].size > 1 ? 'args.inspect' : 'args[0]'}
        $ruby$ language plruby immutable;
      SQL
      unless sql[1].size > 1 or name == 'binary'
        execute <<-SQL
          create or replace function xenodot.#{name}_value(raw bytea) returns #{sql[1][0]} as $ruby$
            #{name == 'long' ? 'raw.to_i' : 'raw'}
          $ruby$ language plruby immutable;
        SQL
      end
    end

    # ruby-postgres seems to have bug sending as time object.
    execute <<-SQL
      create or replace function xenodot.value_date_by_string(text) returns xenodot.value as $sql$
        select * from xenodot.value_date(cast($1 as timestamp with time zone));
      $sql$ language sql;
    SQL
    
    $REFERENCED_TYPES.each do |name, sql|
      execute <<-SQL
        create or replace function xenodot.value_#{name}(val #{sql[1].gsub(/\_head\z/, '')}) returns xenodot.value as $pgsql$
          declare
            result xenodot.value;
          begin
            result.type     := '#{sql[0]}';
            result.content  := (xenodot.value_long(val.id)).content;
            return result;
          end;
        $pgsql$ language plpgsql immutable;
      SQL
        
      execute <<-SQL
        create index #{name}_ref on xenodot.property_head(xenodot.long_value((value).content)) where (value).type = '#{sql[0]}'::char;
      SQL
      
      execute <<-SQL
        create or replace function xenodot.trigger_reference_del_#{name}() returns trigger as $pgsql$
          begin
            perform * from xenodot.property where (value).type = '#{sql[0]}'::char and xenodot.long_value((value).content) = OLD.id;
            if found then
              raise exception '#{sql[1]} still referenced by property, remove first!';
            end if;
            return OLD;
          end;
        $pgsql$ language plpgsql;

        create trigger reference_del_#{name} before delete on #{sql[1]} for each row execute procedure xenodot.trigger_reference_del_#{name}();
      SQL
      
      execute <<-SQL
        create or replace function xenodot.trigger_reference_update_#{name}() returns trigger as $pgsql$
          begin
            if NEW.id <> OLD.id then
              perform * from xenodot.property where (value).type = '#{sql[0]}'::char and xenodot.long_value((value).content) = OLD.id;
              if found then
                raise exception '#{sql[1]} still referenced by property, remove first!';
              end if;
            end if;
            return NEW;
          end;
        $pgsql$ language plpgsql;

        create trigger reference_update_#{name} before update on #{sql[1]} for each row execute procedure xenodot.trigger_reference_update_#{name}();
      SQL

      execute <<-SQL
        create or replace function xenodot.trigger_property_reference_#{name}() returns trigger as $pgsql$
          begin
            if (NEW.value).type = '#{sql[0]}'::char then
              perform * from #{sql[1]} where id = xenodot.long_value((NEW.value).content);
              if not found then
                raise exception '#{sql[1]}(%) does not exist!', xenodot.long_value((NEW.value).content);
              end if;
            end if;
            return NEW;
          end
        $pgsql$ language plpgsql;
        
        create trigger property_reference_#{name} before insert or update on xenodot.property_head 
          for each row execute procedure xenodot.trigger_property_reference_#{name}();
      SQL
      
      execute <<-SQL
        create or replace function xenodot.inline_uri(val xenodot.value) returns bytea as $pgsql$
          declare
            temp xenodot.uri;
            result bytea;
          begin
            select * into temp from xenodot.uri where id = xenodot.long_value((val).content);
            result := (xenodot.value_string(temp.value)).content;
            return result;
          end;
        $pgsql$ language plpgsql stable;
      SQL
      
      execute <<-SQL
        create or replace function xenodot.inline_qname(val xenodot.value) returns bytea as $pgsql$
          declare
            temp xenodot.qname;
            result bytea;
          begin
            select * into temp from xenodot.qname where id = xenodot.long_value((val).content);
            result := (xenodot.value_string('{' || temp.ns || '}' || temp.local_name)).content;
            return result;
          end;
        $pgsql$ language plpgsql stable;
      SQL
    end
  end
  
  def self.down
  end
end