class DomSanity < Xenodot::Migration
  
  def self.up
    execute <<-SQL
      create or replace view xenodot.properties as
        select    node_id                                 as node_id,
                  name_id                                 as prop_name_id,
                  qname.ns                                as prop_ns,
                  qname.local_name                        as prop_local_name,
                  position                                as prop_position,
                  (property.value).type                   as value_type,
                  case (property.value).type
                  when xenodot.type_uri() then
                    xenodot.inline_uri(property.value)
                  when xenodot.type_qname() then
                    xenodot.inline_qname(property.value)
                  else
                    (property.value).content
                  end                                     as value_content
        from      xenodot.property,
                  xenodot.qname
        where     property.name_id = qname.id
        order by  node_id, name_id, position;
    SQL

    execute <<-SQL
      create or replace view xenodot.dom as
        select    node.root_id,
                  node.id                                 as node_id,
                  node.hierarchy                          as node_hierarchy,
                  node_name.id                            as node_name_id,
                  node_name.ns                            as node_ns,
                  node_name.local_name                    as node_local_name,
                  prop_name.id                            as prop_name_id,
                  prop_name.ns                            as prop_ns,
                  prop_name.local_name                    as prop_local_name,
                  prop.position                           as prop_position,
                  (prop.value).type                       as value_type,
                  case (prop.value).type
                  when xenodot.type_uri() then
                    xenodot.inline_uri(prop.value)
                  when xenodot.type_qname() then
                    xenodot.inline_qname(prop.value)
                  else
                    (prop.value).content
                  end                                     as value_content
        from      xenodot.node                            as node
                  join xenodot.qname                      as node_name  on (node_name.id = node.name_id)
                  left join xenodot.property              as prop       on (node.id = prop.node_id)
                  left join xenodot.qname                 as prop_name  on (prop_name.id = prop.name_id) 
        order by  node.root_id, node.hierarchy, prop.name_id, prop.position;
    SQL
    
    execute <<-SQL
      create or replace view xenodot.ns_dom as
        select    node.root_id,
                  node.id                                 as node_id,
                  node.hierarchy                          as node_hierarchy,
                  node_name.id                            as node_name_id,
                  node_ns.prefix                          as node_prefix,
                  node_name.ns                            as node_ns,
                  node_name.local_name                    as node_local_name,
                  prop_name.id                            as prop_name_id,
                  prop_ns.prefix                          as prop_prefix,
                  prop_name.ns                            as prop_ns,
                  prop_name.local_name                    as prop_local_name,
                  prop.position                           as prop_position,
                  (prop.value).type                       as value_type,
                  case (prop.value).type
                  when xenodot.type_uri() then
                    xenodot.inline_uri(prop.value)
                  when xenodot.type_qname() then
                    xenodot.inline_qname(prop.value)
                  else
                    (prop.value).content
                  end                                     as value_content
        from      xenodot.node                            as node
                  join xenodot.qname                      as node_name  on (node_name.id = node.name_id)
                  left join xenodot.uri                   as node_ns    on (node_name.ns = node_ns.value)
                  left join xenodot.property              as prop       on (node.id = prop.node_id)
                  left join xenodot.qname                 as prop_name  on (prop_name.id = prop.name_id) 
                  left join xenodot.uri                   as prop_ns    on (prop_name.ns = prop_ns.value)
        order by  node.root_id, node.hierarchy, prop.name_id, prop.position;
    SQL
    
    
    execute <<-SQL
      create or replace function xenodot.trigger_node_collision() returns trigger as $pgsql$
        declare
          collision xenodot.node;
        begin
          if NEW.hierarchy = '' then
            return NEW;
          end if;

          select * into collision from xenodot.node_head where id <> NEW.id and root_id = NEW.root_id and hierarchy = NEW.hierarchy;
          if found then
            update    xenodot.node_head
              set     hierarchy = xenodot.child_hierarchy(xenodot.parent_hierarchy(hierarchy), xenodot.child_position(hierarchy) + 1)
            where     id = collision.id;
          end if;
          return NEW;
        end;
      $pgsql$ language plpgsql;

      create trigger node_collision before insert or update on xenodot.node_head for each row execute procedure xenodot.trigger_node_collision(); 
    SQL
    
    execute <<-SQL
      create or replace function xenodot.trigger_node_move_descendants() returns trigger as $pgsql$
        declare
          row xenodot.node;
        begin
          if OLD.hierarchy <> NEW.hierarchy or OLD.root_id <> NEW.root_id then
            -- TODO: yes a single update statement is more obvious, but didn't work, trigger order?
            for row in
              select *
              from      xenodot.node_head
              where     root_id = OLD.root_id
                    and xenodot.parent_hierarchy(hierarchy) = OLD.hierarchy
            loop
              update    xenodot.node_head
                set     root_id = NEW.root_id,
                        hierarchy = xenodot.child_hierarchy(NEW.hierarchy, xenodot.child_position(hierarchy))
              where     id = row.id;
            end loop;
          end if;
          return NEW;
        end;
      $pgsql$ language plpgsql;

      create trigger node_move_descendants after update on xenodot.node_head for each row execute procedure xenodot.trigger_node_move_descendants();
    SQL
    
    execute %q{
      create or replace function xenodot.create_array(xenodot.node, xenodot.qname) returns xenodot.property as $pgsql$
        declare
          result xenodot.property;
        begin
          insert into xenodot.property_head(node_id, name_id, position, value) values ($1.id, $2.id, 0, (null, null)) returning * into result;
          return result;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute <<-SQL
      create or replace function xenodot.trigger_property_array() returns trigger as $pgsql$
        declare
          parent_id int4;
          value_type "char";
          prop_name xenodot.qname;
        begin
          if NEW.position < 1 then
            return NEW;
          end if;

          select (value).type into value_type from xenodot.property_head where node_id = NEW.node_id and name_id = NEW.name_id and position = 0;
          if not found then
            -- lock the node
            select id into parent_id from xenodot.node_head where id = NEW.node_id for share;

            -- try reading again
            select (value).type into value_type from xenodot.property_head where node_id = NEW.node_id and name_id = NEW.name_id and position = 0;

            if not found then
              select * into prop_name from xenodot.qname where id = NEW.name_id;
              select (value).type into value_type from xenodot.create_array(xenodot.node(NEW.node_id), prop_name);
            end if;
          end if;

          if value_type is not null then
            raise exception 'Attribute is not an array! %', value_type;
          end if;

          if NEW.position is null then
            select id into parent_id from xenodot.node_head where id = NEW.node_id for share;
            select count(*) into NEW.position from xenodot.property_head where node_id = NEW.node_id and name_id = NEW.name_id;
          end if;

          return NEW;
        end;
      $pgsql$ language plpgsql;

      create trigger property_array before insert or update on xenodot.property_head for each row execute procedure xenodot.trigger_property_array();
    SQL
  end

  def self.down
  end

end