class Xpath < Xenodot::Migration

  def self.up
    execute <<-SQL
      insert into xenodot.qname(id, ns, local_name) values(-2, '!xenodot!', 'xml-text'), (-3, '!xenodot!', 'xml-attr');
      
      create or replace function xenodot.xml_meta_name() returns xenodot.qname as $sql$
        select * from xenodot.qname where id = -2;
      $sql$ language sql immutable;
      
      create or replace function xenodot.xml_text_attr_name() returns xenodot.qname as $sql$
        select * from xenodot.qname where id = -3;
      $sql$ language sql immutable;
    SQL

    execute %q{
      create or replace function xpath.string(root xenodot.node) returns text as $ruby$
        query = $Plans[:string] ||= PL::Plan.new(%q{
          select    (property.value).content
          from      xenodot.node,
                    xenodot.property,
                    xenodot.xml_meta_name() as node_name,
                    xenodot.xml_text_attr_name() as prop_name
          where     node.root_id = $1
                and node.hierarchy like $2
                and node.name_id = node_name.id
                and property.node_id = node.id
                and property.name_id = prop_name.id
          order by node.hierarchy
        }, ['int4', 'text']).save
        query.exec([root['root_id'], root['hierarchy'] + '%'], 'output' => 'value').join
      $ruby$ language plruby stable;
    }
    
    # TODO: URIs and QNames are double encoded
    execute %q{
      create or replace function xpath.string(xenodot.property) returns text as $sql$
        select xenodot.string_value(($1.value).content);
      $sql$ language sql stable;
    }
    
    execute %q{
      create or replace function xpath.node_name(xenodot.node) returns text as $sql$
        select  '{' || qname.ns || '}' || qname.local_name
        from    xenodot.qname
        where   id = $1.name_id;
      $sql$ language sql stable;
    }
    
    execute %q{
      create or replace function xpath.contains(text, text) returns boolean as $sql$
        select position($2 in $1) > 0;
      $sql$ language sql immutable;
    }
    
    execute %q{
      create or replace function xenodot.node_cursor(xenodot.node, text) returns refcursor as $ruby$
        init_xenodot
        xpath_query = Xenodot::Query::XPath.build_query(*args)
        xpath_query.sql?.default = 'current_node.*'
        xpath_query.sql!(
          :cursor => true,
          :order_by => 'current_node.root_id, current_node.hierarchy'
        )
      $ruby$ language plruby;
    }
    
    execute %q{
      create or replace function xpath.node(xenodot.node, text) returns setof xenodot.node as $pgsql$
        declare
          ref refcursor;
          row xenodot.node;
        begin
          ref := xenodot.node_cursor($1, $2);
          loop
            fetch ref into row;
            if not found then
              close ref;
              return;
            end if;
            return next row;
          end loop;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.dom_cursor(xenodot.node, text) returns refcursor as $ruby$
        init_xenodot
        xpath_query = Xenodot::Query::XPath.build_query(*args)

        xpath_focus = xpath_query.sql?.default.to_s.scan(/\_(\w+)\.\*/).to_s
        case xpath_focus
        when 'property':
          xpath_query = Xenodot::Query::AbstractQuery.new(xpath_query) do |sql|
            sql.from  = 'xenodot.dom'
            sql.where = 'dom.node_id = parent1_node.id and dom.prop_name_id = parent1_property_name.id'
            sql.default = 'dom.*'
          end
        when 'node':
          xpath_query = Xenodot::Query::AbstractQuery.new(xpath_query) do |sql|
            sql.from  = 'xenodot.dom'
            sql.where = 'dom.node_id = parent1_node.id'
            sql.default = 'dom.*'
          end
        else
          raise "Panic on xpath build, got #{xpath_focus}"
        end
        xpath_query.sql!(
          :cursor => true
        )
      $ruby$ language plruby;
    }
    
    execute %q{
      create or replace function xpath.dom(xenodot.node, text) returns setof xenodot.dom as $pgsql$
        declare
          ref refcursor;
          row xenodot.dom;
        begin
          ref := xenodot.dom_cursor($1, $2);
          loop
            fetch ref into row;
            if not found then
              close ref;
              return;
            end if;
            return next row;
          end loop;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.mime_type(root xenodot.node, xpath text) returns text as $ruby$
        init_xenodot
        xpath = "#{xpath}/" unless xpath =~ /\/\z/
        xpath = "#{xpath}ancestor-or-self::/@xen:mime-type"
        query = Xenodot::Query::XPath.build_query(root, xpath)
        query.sql?.default = 'distinct (current_property.value).content'
        candidates = query.sql!
        if canidates
          candidates = candidates.flatten.uniq
          if candidates.size == 1
            return candidates[0] 
          else
            return 'text/xml'
          end
        end
        return nil
      $ruby$ language plruby stable;
      
      create or replace function xenodot.mime_type(node xenodot.node) returns text as $sql$
        select * from xenodot.mime_type($1, '.');
      $sql$ language sql stable;
    }
  end

  def self.down
  end

end