class XmlParser < Xenodot::Migration

  def self.up

    execute %q{
      create or replace function xenodot.get_uri(uri text) returns void as $ruby$
        init_xenodot

        entry = Xenodot::Backend.sql('select root_id, hierarchy from xenodot.clean_root($1)', [uri], [:text], :count => 1)
        Xenodot::Backend::XmlParser.parse(URI.parse(uri), *entry)
      $ruby$ language plruby;
    }
    
    execute %q{
      create or replace function xenodot.append_xml(root xenodot.node, xml text) returns void as $ruby$
        init_xenodot
        Xenodot::Backend::XmlParser.parse(xml, root['root_id'], root['hierarchy'])
      $ruby$ language plruby;
    }
    
    xml_meta_name_id = execute(%q{
      select id from xenodot.xml_meta_name();      
    }).result.to_s.to_i
    
    create_singleton :xml_meta_name_id do
      xml_meta_name_id
    end
    
    xml_text_attr_name_id = execute(%q{
      select id from xenodot.xml_text_attr_name();
    }).result.to_s.to_i
    
    create_singleton :xml_text_attr_name_id do
      xml_text_attr_name_id
    end
    
    create_singleton :quote_xml, :type, :content do %q{
      case type
      when 's':
        content.gsub!('<', '&lt;')
        content.gsub!('>', '&gt;')
        content.gsub!('"', '&quot;')
        content
      else
        content
      end
      
    } end
    
    create_singleton :build_xml, :cursor, :start_row, :result, :namespaces do %q{
      if start_row['node_hierarchy'] == ''
        return cursor.fetch
      elsif start_row['node_name_id'] == xml_meta_name_id()
        if start_row['prop_name_id'] == xml_text_attr_name_id()
          result << quote_xml(start_row['value_type'], start_row['value_content'])
        end
        return cursor.fetch
      end
      
      attribs = {}
      
      curr_namespaces = {}
      if start_row['node_prefix'] != ''
        node_name = "#{start_row['node_prefix']}:#{start_row['node_local_name']}"
        unless namespaces.include?(start_row['node_prefix'])
          curr_namespaces[start_row['node_prefix']] = start_row['node_ns']
        end
      else
        node_name = start_row['node_local_name']
      end
      
      row = start_row
      begin
        if row != nil and row['node_id'] == start_row['node_id']
          if row['prop_prefix'] == ''
            prop_name = row['prop_local_name']
          elsif row['prop_prefix']
            prop_name = "#{row['prop_prefix']}:#{row['prop_local_name']}"
            unless namespaces.include?(row['prop_prefix'])
              curr_namespaces[row['prop_prefix']] = row['prop_ns']
            end
          else
            prop_name = nil
          end
          if prop_name
            if attribs[prop_name]
              attribs[prop_name] << ' ' << quote_xml(row['value_type'], row['value_content'])
            else
              attribs[prop_name] = quote_xml(row['value_type'], row['value_content'])
            end
          end
        else
          result << '<' << node_name
          if curr_namespaces.size > 0
            result << ' ' << curr_namespaces.collect{|prefix, uri| %Q{xmlns:#{prefix}="#{uri}"}}.join(' ')
          end
          if attribs.size > 0
            result << ' ' << attribs.collect{|name, value| %Q{#{name}="#{value}"}}.join(' ')
          end
          result << '>'
          
          begin
            if row.nil? or row['root_id'] != start_row['root_id'] or row['node_hierarchy'].index(start_row['node_hierarchy']) != 0
              result << '</' << node_name << '>'
              return row
            else
              row = build_xml(cursor, row, result, namespaces.merge(curr_namespaces))
            end
          end while true
        end
        row = cursor.fetch
      end while true
    } end
    
    execute %q{
      create or replace function xpath.xml(xenodot.node, text) returns bytea as $ruby$
        init_xenodot
        xpath_query = Xenodot::Query::XPath.build_query(*args)

        xpath_focus = xpath_query.sql?.default.to_s.scan(/\_(\w+)\.\*/).to_s
        case xpath_focus
        when 'property':
          xpath_query = Xenodot::Query::AbstractQuery.new(xpath_query) do |sql|
            sql.from  = 'xenodot.ns_dom'
            sql.where = 'ns_dom.node_id = parent1_node.id and ns_dom.prop_name_id = parent1_property_name.id'
            sql.default = 'ns_dom.*'
          end
        when 'node':
          xpath_query = Xenodot::Query::AbstractQuery.new(xpath_query) do |sql|
            sql.from  = 'xenodot.ns_dom'
            sql.where = 'ns_dom.node_id = parent1_node.id'
            sql.default = 'ns_dom.*'
          end
        else
          raise "Panic on xpath build, got #{xpath_focus}"
        end
        begin
          cursor = xpath_query.sql!(
#            :debug => true,
            :cursor => true,
            :output => 'hashed'
          )

          result = ''
          row = cursor.fetch
          while row
            row = build_xml(cursor, row, result, {})
          end
          return result
        ensure
          cursor.close
        end
      $ruby$ language plruby stable;
    }
    
    execute %q{
      create or replace function xenodot.to_xml(xenodot.node) returns bytea as $sql$
        select xpath.xml($1, $xpath$ descendant-or-self:: $xpath$);
      $sql$ language sql stable;
    }

  end

  def self.down
  end

end