class UriTemplating < Xenodot::Migration

  def self.up
    execute %q%
      create or replace function xenodot.generate_uri(template text) returns xenodot.uri as $ruby$
        init_xenodot
        
        # make sure the template contains '{uniq}'
        template.gsub!(/\Z/, '{uniq}') unless template =~ /\{uniq\}/

        # the regex to find the template
        regexp = Regexp.new('(\/?)\{uniq\}(\/?)')
        
        # now check the clean version for authority
        clean = template.gsub(regexp, '/')
        return nil unless Xenodot::Backend.sql('select xenodot.uri_authority($1)', [clean], [:text], :count => 1)[0]
        
        now = Time.now
        uri = [
          clean,
          template.gsub(regexp, "/#{now.year}/#{now.month}/#{now.day}/"),
          template.gsub(regexp, "/#{now.year}/#{now.month}/#{now.day}/\{uniq\}/")
        ].find do |candidate|
          candidate.gsub!(/\{uniq\}/) do
            Xenodot::Backend.sql("select * from nextval('xenodot.uri_uniquer');", [], []).to_s
          end
          Xenodot::Backend.sql('select * from xenodot.uri($1)', [candidate], [:text], :count => 1).compact.size == 0
        end
        Xenodot::Backend.sql('select * from xenodot.ensure_uri($1)', [uri], [:text], :count => 1)
      $ruby$ language plruby;
    %
    
    execute <<-SQL
      create sequence xenodot.uri_uniquer;
    SQL
    
    create_table 'xenodot.uri_authority' do |t|
      t.column :value,  :text, :null => false
    end
    
    execute <<-SQL
      create index hierarchy on xenodot.uri_authority(xenodot.uri_hierarchy(value));
    SQL
    
    hostname = %x{hostname}.strip
    execute <<-SQL
      insert into xenodot.uri_authority(value) values ('http://#{hostname}:3000'), ('http://localhost:3000');
    SQL
    
    execute <<-SQL
      create or replace function xenodot.uri_authority(test text) returns boolean as $pgsql$
        declare
          result boolean;
        begin
          select  count(*) > 0 
          into    result
          from    xenodot.uri_authority 
          where   xenodot.uri_hierarchy(test) like xenodot.uri_hierarchy(value) || '%';
          
          return result;
        end;
      $pgsql$ language plpgsql stable;
    SQL
    
    ['', 'ensure_'].each do |prefix|
      stable = prefix == '' ? 'stable' : ''
      
      execute <<-SQL
        create or replace function xenodot.#{prefix}root_uri(uri text) returns xenodot.uri as $pgsql$
          declare
            result xenodot.uri;
          begin
            if xenodot.uri_authority(uri) then
              select * into result from xenodot.#{prefix}uri(substring(uri from '.*://(.+)|.+'));
            else
              select * into result from xenodot.#{prefix}uri(uri);
            end if;
            return result;
          end;
        $pgsql$ language plpgsql #{stable};
      SQL

      execute <<-SQL
        create or replace function xenodot.#{prefix}root_id(uri text) returns int4 as $sql$
          select id from xenodot.#{prefix}root_uri($1);
        $sql$ language sql #{stable};
      SQL
    end

  end
  
  def self.down
  end

end