class LastModified < Xenodot::Migration
  def self.up
    create_table 'xenodot.last_modified', :id => false do |t|
      t.column :root_id,    :int4, :null => false
      t.column :hierarchy,  :text, :null => false
      t.column :node_id,    :int4, :null => false
      t.column :prop_id,    :int4, :null => false
      t.column :stamp,      'timestamp with time zone'
    end
    
    execute %q{
      alter table xenodot.last_modified add foreign key(node_id) references xenodot.node_head on delete cascade initially deferred;
      alter table xenodot.last_modified add foreign key(prop_id) references xenodot.property_head on delete cascade initially deferred;
      create index tree_idx on xenodot.last_modified(root_id, hierarchy);
      create index node_idx on xenodot.last_modified(node_id);
      create index prop_idx on xenodot.last_modified(prop_id);
      create index time_idx on xenodot.last_modified(stamp);
    }
    
    execute %q{
      create rule node_insert as
        on insert to xenodot.node
      do instead (
        insert into xenodot.node_head(root_id, hierarchy, name_id) values(NEW.root_id, NEW.hierarchy, NEW.name_id) returning *;
      );
    }
    
    execute %q{
      create rule node_update as
        on update to xenodot.node
      do instead (
        update    xenodot.node_head
          set     root_id = NEW.root_id,
                  hierarchy = NEW.hierarchy,
                  name_id = NEW.name_id
        where     id = OLD.id returning *;
      );
    }
    
    execute %q{
      create rule node_delete as
        on delete to xenodot.node
      do instead(
        delete    
        from      xenodot.node_head
        where     root_id = OLD.root_id
              and hierarchy like OLD.hierarchy || '%'
        returning *
      );
    }
    
    execute %q{
      create rule prop_insert as
        on insert to xenodot.property
      do instead(
        insert into xenodot.property_head(node_id, name_id, position, value) values(NEW.node_id, NEW.name_id, NEW.position, NEW.value) returning *;
      );
    }
    
    execute %q{
      create rule prop_update as
        on update to xenodot.property
      do instead(
        update    xenodot.property_head
          set     name_id = NEW.name_id,
                  position = NEW.position,
                  value = NEW.value
        where     id = NEW.id
        returning *;
      );
    }
    
    execute %q{
      create rule prop_delete as
        on delete to xenodot.property
      do instead(
        delete
        from      xenodot.property_head
        where     id = OLD.id
        returning *;
      );
    }
    
    execute %q{
      create or replace function xenodot.last_modified(xenodot.node) returns timestamp with time zone as $sql$
        select    stamp
        from      xenodot.last_modified
        where     root_id = $1.root_id
              and hierarchy = any(xenodot.ancestor_hierarchy($1.hierarchy))
        limit 1;
      $sql$ language sql stable;
    }
    
    execute %q{
      create or replace function xenodot.last_modified(root xenodot.node, xpath text) returns timestamp with time zone as $ruby$
        init_xenodot
        query = Xenodot::Query::AbstractQuery.new(Xenodot::Query::XPath.build_query(root, xpath)) do |sql|
          sql.from = 'xenodot.last_modified'
          sql.default = 'stamp'
          sql.where = %q{ last_modified.root_id = parent1_node.root_id
                      and last_modified.hierarchy = any(xenodot.ancestor_hierarchy(parent1_node.hierarchy))
          }
        end
        result = query.sql!(:count => 1)
        return result ? result[0] : null
      $ruby$ language plruby stable;  
    }
    
    

  end
  
  def self.down
  end
end