require 'active_record'

ActiveRecord::Migration.verbose = $XENODOT_DEBUG

module Xenodot
  SCHEMAS = %w{xenodot xpath jcr}
  
  class Migration < ActiveRecord::Migration
    def self.create_singleton(name, *args)
      execute <<-SQL
        insert into plruby_singleton_methods(name, args, body) values('#{name}', '#{args.join(',')}', $ruby$#{yield}$ruby$)
      SQL
    end 
  end
end

module Kernel
  def reset_xenodot_database
    begin
      ActiveRecord::Base.remove_connection
      Xenodot::Backend::Database.remove_connection
    rescue
    end
    clean_xenodot_database
    connect_xenodot_database
    migrate_xenodot_database
  end
  
  def clean_xenodot_database
    spec = YAML.load_file(File.join(File.dirname(__FILE__),'..','config','database.yml'))[$XENODOT_ENV]
    retry_counter = 0
    begin
      connect_xenodot_database
      Xenodot::Backend.sql('set client_min_messages to error;', [], [])
    rescue  => ex
      retry_counter += 1
      %x{createdb -q -h #{spec['host']} -U #{spec['username']} -E utf8 #{spec['database']}}
      retry unless retry_counter > 2
      puts "Failed to createdb:"
      puts ex
    end
    
    Xenodot::SCHEMAS.each do |schema|
      Xenodot::Backend.sql("drop schema #{schema} cascade;", [], []) rescue {}
    end
    Xenodot::Backend.sql('drop table plruby_singleton_methods;', [], []) rescue {}
    
    begin
      Xenodot::Backend.sql(File.new(File.join(File.dirname(__FILE__), 'uninstall_tsearch2.sql')).read, [], [])
      Xenodot::Backend.sql('select 1;', [], []) # test if removing tsearch2 succeeded
    rescue
      Xenodot::Backend.sql('rollback;', [], [])
    end
    Xenodot::QName.clear_cache
  end
  
  def migrate_xenodot_database(version = nil)
    Xenodot::Backend.sql('set client_min_messages to error;', [], []) unless $XENODOT_DEBUG
    # the pljava.sql create the schema itself
    Xenodot::SCHEMAS.each do |schema|
      Xenodot::Backend.sql("create schema #{schema};", [], []) rescue {}
    end
    ActiveRecord::Migrator.migrate(File.join(File.dirname(__FILE__),"migration"), version)
  end
  
  def connect_xenodot_database
    ActiveRecord::Base.logger = Logger.new($XENODOT_DEBUG ? STDOUT : NIL)
    Xenodot::Backend::Database.connect!(ActiveRecord::Base)
    ActiveRecord::Base.connection.execute 'set client_min_messages to error;'
  end
end