module Xenodot
  
  def self.encode(val)
    case val
    when String, Symbol, URI:
      "E'#{val.to_s.gsub(/(['\\])/){ "\\#{$1}" }}'"
    else
      val.to_s
    end
  end
  
  class MinimizedSlate
    instance_methods.each do |m| 
      unless m =~ /^__|^respond_to?|^inspect|^to_s|^kind_of?|^class|^should|^instance_of?|^kind_of?|object_id|equals?|^send|^nil|^method|^_method_missing/
        undef_method m
      end
    end
  end

  module MetaAid
    def metaclass; class << self; self; end; end

    alias act_as extend

    def instance_exec &block # This is instance_exec
      method_name = "Context_#{Time.now.to_i}_#{Time.now.usec}"
      metaclass.send(:define_method, method_name, block)
      begin
        self.send(method_name)
      ensure
        metaclass.send(:remove_method, method_name)
      end
    end
  end
  
  class QName
  
  	@@cache = {}
  	
  	def self.clear_cache
  		@@cache.clear
  	end
  	
  	def self.id(*args)
        ns = args[-2] || ''
        local_name = args[-1].to_s

  		id = @@cache["{#{ns.to_s}}#{local_name}"]
  		id || QName[*args].id
  	end
		
    def self.[](*args)
      raise 'Expected one or two parameters' unless args.size.between?(1,2)
      params = if args[0].kind_of? Integer
        Backend.sql('select ns, local_name, id from xenodot.qname where id = $1', [args[0]], [:int4], :count => 1)
      else
        ns = args[-2] || ''
        local_name = args[-1]
        [ns.to_s, local_name.to_s]
        #Backend.sql('select ns, local_name, id from xenodot.ensure_qname($1, $2)', [ns.to_s, local_name.to_s], [:text, :text], :count => 1)
      end
      new(*params)
    end
	  
    def initialize(ns, local_name, id = nil)
      @ns = ns.kind_of?(URI) ? ns : URI.parse(ns)
      @local_name = local_name
      @id = id
	  @@cache[self.to_s] = id if id
    end
    
    def id
      unless @id
      	@id = @@cache[self.to_s]
      	unless @id
			@id = Backend.sql('select id from xenodot.ensure_qname($1, $2)', [@ns.to_s, @local_name], [:text, :text], :count => 1)[0]
			@@cache[self.to_s] = @id	
		end
      end
      @id
    end
    
    def to_s
      "{#{@ns}}#{@local_name}"
    end
    
    def sql_value
      if @id
        "xenodot.value_qname(xenodot.qname(#{Xenodot.encode(@ns)}, #{Xenodot.encode(@local_name)}))"
      else
        "xenodot.value_qname(xenodot.ensure_qname(#{Xenodot.encode(@ns)}, #{Xenodot.encode(@local_name)}))"
      end
    end
    
    def ns
      @ns
    end
    
    def hash
      id
    end
    
    def eql?(other)
      other.kind_of? Xenodot::QName and other.ns == self.ns and other.local_name == self.local_name
    end
    
    alias == eql?

    attr_reader :local_name
  end
end

module URI
  
  def self.generate_from(template)
    uri = Xenodot::Backend.sql('select value from xenodot.generate_uri($1)', [template], [:text], :count => 1)[0]
    return uri.nil? ? nil : URI.parse(uri)
  end
  
  def self.by_prefix(prefix)
    URI.parse(Xenodot::Backend.sql('select value from xenodot.uri($1)', [prefix], [:text], :count => 1)[0])
  end
  
  def self.by_id(id)
    URI.parse(Xenodot::Backend.sql('select value from xenodot.uri where id = $1', [id], [:int4], :count => 1)[0])
  end
  
  def self.exist?(uri)
    Xenodot::Backend.sql('select * from xenodot.uri($1)', [uri.to_s], [:text], :count => 1).compact.size > 0
  end
  
  def self.authorative?(uri)
    Xenodot::Backend.sql('select xenodot.uri_authority($1)', [uri.to_s], [:text], :count => 1)[0]
  end
  
  def self.PUT(request, response)
    uri = request.params['REQUEST_URI']
    if URI.exist?(uri) and URI.authorative?(uri)
      puts "AUTHORITIVE"
      root = Xenodot::Node[uri].clean!
      puts root[:content_type] = request.params['CONTENT-TYPE'] || 'text/xml'
      root.append_xml(request.body.read)
      
      response.start(200) do |head, out|
        out.write Xenodot::Node[uri].to_xml
      end
    else
      puts "REFUSE"
      response.start(403) do |head, out|
        out.write 'Forbidden, not authoritive!'
      end
    end
  end
    
  class Generic
    def id
      @id ||= Xenodot::Backend.sql('select id from xenodot.ensure_uri($1)', [self.to_s], [:text], :count => 1)[0]
    end
    
    def prefix
      Xenodot::Backend.sql('select prefix from xenodot.ensure_uri($1)', [self.to_s], [:text], :count => 1)[0]
    end
    
    def prefix=(new_prefix)
      if self.prefix != new_prefix # ensures existance
        Xenodot::Backend.sql('update xenodot.uri set prefix = $2 where value = $1', [self.to_s, new_prefix.to_s], [:text, :text])
      end
    end
    
    def sql_value
      if @id
        "xenodot.value_uri(xenodot.uri(#{Xenodot.encode(to_s)}))"
      else
        "xenodot.value_uri(xenodot.ensure_uri(#{Xenodot.encode(to_s)}))"
      end
    end
  end
end

class Time
  def sql_value
    "xenodot.value_date_by_string(#{Xenodot.encode(rfc822)})"
  end
end

class Integer
  def sql_value
    "xenodot.value_long(#{to_s})"
  end
end

class Float
  def sql_value
    "xenodot.value_double(#{to_s})"
  end
end

class String
  def sql_value
    "xenodot.value_string(#{Xenodot.encode(self)})"
  end
end

class TrueClass
  def sql_value
    'xenodot.value_boolean(true)'
  end
end

class FalseClass
  def sql_value
    'xenodot.value_boolean(false)'
  end
end

class Symbol
  def sql_value
    "xenodot.value_string(#{Xenodot.encode(to_s)})"
  end
end