module Xenodot
  class Node < MinimizedSlate
    include NodeInstanceMethods

    def initialize(id)
      raise "Illegal node id #{id.inspect}" unless id.kind_of? Integer
      @id = id
    end
    
    def id(options = {:array => false})
      options[:array] ? [@id] : @id
    end
    
    def start_node_id
      @id
    end
    
    def xpath
      '.'
    end
    
    def self.[](uri_or_id)
      if uri_or_id.kind_of? Integer
        new(Backend.sql('select id from xenodot.node($1)', [uri_or_id], [:int4], :count =>  1)[0])
      else
        new(Backend.sql('select id from xenodot.root_node($1)', [uri_or_id.to_s], [:text], :count => 1)[0])
      end
    end
    
    def == (other)
      if other.kind_of? Node
        self.id == other.id
      elsif other.respond_to? :materialize!
        self == other.materialize!
      else
        false
      end
    end
    
    def hash
      self.id
    end
    
    def exist?
      Backend.sql('select 1 from xenodot.node where id = $1',[self.id],[:int4],:count => 1) != false
    end
    
    def size
      return exist? ? 1 : 0
    end
    
    def delete!(options = {:array => false})
      deleted = Backend.sql('delete from xenodot.node where id = $1 returning *', [self.id], [:int4]) || []
      deleted = deleted[0] unless deleted.size > 1 and options[:array]
      return deleted
    end
    
    def materialize!(options = {:array => false})
      options[:array] ? [self] : self
    end
    
    def parent=(new_parent)
      raise "Must be a node" unless new_parent.kind_of? Node
      Backend.sql('select xenodot.move_to(xenodot.node($1), xenodot.node($2))', [self.id, new_parent.id], [:int4, :int4])
    end
    
    def sql_value
      "xenodot.value_node(xenodot.node(#{@id}))"
    end

  end
end