module Xenodot
  module NodeInstanceMethods
  
  	include MetaAid
  	include Enumerable
  
    %w{child parent
       ancestor ancestor_or_self descendant
       descendant_or_self
       following following_sibling
       preceding preceding_sibling
    }.each do |accessor|      
      eval %Q{
        def #{accessor}(*args)
          name_params, options = NodeInstanceMethods.parse_qname_and_options(args, :allow_empty => true, :xpath => true)
          xpath_accessor = (['#{accessor.gsub(/\_/, '-')}:'] + name_params).join(':')
          Xenodot::XPathBuilder.new(self, "\#{xpath_accessor}")
        end
      }
    end
    
    def to_xml
      xpath_builder = Xenodot::XPathBuilder.new(self, 'descendant-or-self::')
      Backend.sql('select xpath.xml(xenodot.node($1), $2)', [xpath_builder.start_node_id, xpath_builder.xpath], [:int4, :text], :count => 1)[0]
    end
    
    alias to_s to_xml
    
    def name(options = {:array => false})
      result = if self.kind_of? Node
			Backend.sql(%q{
			select  qname.ns, qname.local_name, qname.id 
			from    xenodot.qname, xenodot.node($1)
			where   qname.id = node.name_id
			}, [start_node_id], [:int4])
      	else
			Backend.sql(%q{
			select  qname.ns, qname.local_name, qname.id 
			from    xenodot.qname, xpath.node(xenodot.node($1), $2)
			where   qname.id = node.name_id
			}, [start_node_id, xpath], [:int4, :text])
	  end
      if result
        result = result.collect{|row| Xenodot::QName.new(*row)}
        result = result[0] unless result.size > 1 or options[:array]
      end
      result
    end
    
    def clean!
      Xenodot::Backend.sql(%Q{
        delete from xenodot.property where node_id in (#{id(:array => true).join(',')});
        
        delete from xenodot.node where id in (
          select    descendant.id 
          from      xenodot.node as root, xenodot.node as descendant
          where     root.id in (#{id(:array => true).join(',')})
                and descendant.root_id = root.root_id
                and descendant.hierarchy like root.hierarchy || '%_'
        )
      }, [], [])
      self
    end
    
    # def create_child, def ensure_child
    %w{create ensure}.each do |meth|
      eval %Q{
        def #{meth}_child(*args)
          name_params, options = NodeInstanceMethods.parse_qname_and_options(args)
          children = Xenodot::Backend.sql(%q{
            select  (xenodot.#{meth}_child(node, qname)).id
            from    xenodot.ensure_qname($1, $2) as qname, xpath.node(xenodot.node($3), $4)
          }, name_params + [start_node_id, xpath], [:text, :text, :int4, :text])
          if children
            children = children.collect{|row| Node[row[0]]}
            children = children[0] unless children.size > 1 or options[:array]
          end
          children
        end
      }
    end
    
    def append_xml(content)
      Xenodot::Backend.sql('select xenodot.append_xml(node, $1) from xenodot.node where id in (' << id(:array => true).join(',') << ")", [content], [:text])
      self
    end
    
    def append_clone(sources, options = {:array => false})
      result = sources.materialize!(:array => true).collect do |source|
        Node.new(*Backend.sql(
          'select (xenodot.clone_to(xenodot.node($1), node)).id from xpath.node(xenodot.node($2), $3);',
          [source.id, start_node_id, xpath],
          [:int4, :int4, :text],
          :count => 1
        ))
      end
      result = result[0] unless result.size > 1 or options[:array]
      return result
    end
    
    def transaction! &block
      return unless block
      Xenodot::Backend.transaction! &block
    end
    
    def attributes
      db_result = Backend.sql(
        'select distinct prop_ns, prop_local_name, prop_name_id from xpath.dom(xenodot.node($1), $2)',
        [start_node_id, xpath], [:int4, :text]
      )
      if db_result.size == 1 and db_result[0][0].nil?
        []
      else
        db_result.collect{|row| QName.new(*row)}
      end
    end
    
    def [](*args)
      if  self.kind_of? Xenodot::XPathBuilder and args.size == 1
        case args[0]
        when Integer:
          pos = args[0]
          if pos < 0
            return XPathBuilder.new(start_node_id, xpath, "rposition() = #{pos}")
          else
            return XPathBuilder.new(start_node_id, xpath, "position() = #{pos+1}")
          end
        when Range:
          condition = ''
          first = args[0].first
          if first < 0
            condition << "rposition() >= #{first}"
          else
            condition << "position() > #{first}"
          end
          condition << ' and '
          last = args[0].last
          if last < 0
            last -= 1 if args[0].exclude_end?
            condition << "rposition() <= #{last}"
          else
            last += 1 unless args[0].exclude_end?
            condition << "position() <= #{last}"
          end
          return XPathBuilder.new(start_node_id, xpath, condition)
        end
      end
      name_params, options = NodeInstanceMethods.parse_qname_and_options(args)
		name_id = QName.id(*name_params)
		db_result = if self.kind_of? Node
			Backend.sql(%q{
				select properties.node_id, properties.prop_position, properties.value_type, properties.value_content
				from   xenodot.node($1) as x, xenodot.properties
				where     properties.node_id = x.id
					and properties.prop_name_id = $2 order by properties.prop_position
			}, [self.start_node_id, name_id], [:int4, :int4], :output => :hashed)
		else
			Backend.sql(%q{
				select properties.node_id, properties.prop_position, properties.value_type, properties.value_content
				from   xpath.node(xenodot.node($1), $2) as x, xenodot.properties
				where     properties.node_id = x.id
					and properties.prop_name_id = $3 order by properties.prop_position
			}, [self.start_node_id, self.xpath, name_id], [:int4, :text, :int4], :output => :hashed)
		end

      return options[:array] ? [] : nil unless db_result
      current_node_id = nil
      result = []
      db_result.each do |row|
        if current_node_id != row['node_id']
          pos = result.size
          current_node_id = row['node_id']
        else
          pos = -1
        end
        curr_val = case row['value_type']
        when 's': row['value_content']
        when 'u': URI.parse(row['value_content'])
        when 'q': QName[*row['value_content'].scan(/\{(.+)\}(.+)/).flatten]
        when 'n': 
          Node.new(row['value_content'].to_i)
        #when 'b': Todo
        when 'p': 
        	x = row['value_content'].split(/\|/,2)
        	XPathBuilder.new(x[0].to_i,x[1])
        when 't': Time.local(*ParseDate.parsedate(row['value_content']))
        when 'd': row['value_content'].to_f
        when 'l': row['value_content'].to_i
        when '1': 'true' == row['value_content']
        when nil: []
        else
          raise "Panic, unknown value_type #{row['value_type']}:#{row['value_content'].inspect}"
        end
        if row['prop_position'] == 0
          result[pos] = curr_val
        else
          result[pos] << curr_val
        end
      end
      result = result[0] unless result.size > 1 or options[:array]
      return result
    end
    
    def []=(*args)
      name_params, options, value = NodeInstanceMethods.parse_qname_and_options(args, :has_value => true)
      
      transaction! do
        name_id = nil
        name_id = QName.id(*name_params)
        node_ids = id(:array => true)
        

        # remove existing properties
        Backend.sql(
          "delete from xenodot.property where node_id in (#{node_ids.join(',')}) and name_id = #{name_id}", [], []
        )
        
        unless value.nil? # setting nil is removing
          value = case value
          when Array
            value.flatten.collect{|array_val| array_val.sql_value}.flatten
          else
            value.sql_value
          end
          
          insert_values = node_ids.collect do |node_id|
            case value
            when Array:
              if value.size > 0
                array_offset = 0
                value.collect do |sql_value|
                  "(#{node_id}, #{name_id}, #{array_offset+=1}, #{sql_value})"
                end
              else
                "(#{node_id}, #{name_id}, 0, (null, null))" # array header
              end
            else
              "(#{node_id}, #{name_id}, 0, #{value})"
            end
          end.flatten
          Backend.sql(
            "insert into xenodot.property(node_id, name_id, position, value) values #{insert_values.join(',')} returning *", [], []
          )
        end
      end
    end

    def transaction?
      Xenodot::Backend.transaction?
    end
    
    def sql_value
      "xenodot.value_path(#{Xenodot.encode(self.xpath)})"
    end
    
    # TODO make it better
    def write(other)
    	self.append_xml(other.to_xml)
    end
    
    def <=>(other)
    	self.id <=> other.id
    end
    
    def each
    	materialize!(:array => true).each{|i| yield i}
    end
    
    def self.parse_qname_and_options(args, options={})
      call_options = args[-1].kind_of?(Hash) ? args.pop : {:array => false}
      call_value = args.pop if options[:has_value]
      
      if args.size == 0 and options[:allow_empty]
        args = [''] if options[:xpath]
      else
        raise "Syntax error #{args.inspect}" unless args.size.between?(1,2)
        if args[0].kind_of? Xenodot::QName
          if args.size == 1
            args = [args[0].ns.prefix, args[0].local_name]
          else
            args = [args[0].ns.prefix, args[1]]
          end
        elsif args.size == 2 and args[0].kind_of? URI
          args[0] = args[0].prefix
        end
        args = args.collect{|n| n.to_s}
        args = ['', args[0]] if args.size == 1 and not options[:xpath]
      end
      
      return args, call_options, call_value
    end
    
	def fast_child(*args)
		name_params, options = NodeInstanceMethods.parse_qname_and_options(args, :allow_empty => true)
		id = self.id
		res = unless name_params.empty?
			name_id = QName.id(*name_params)
			Backend.sql('select id from xenodot.node where (root_id, xenodot.parent_hierarchy(hierarchy)) = (select root_id, hierarchy from xenodot.node($1)) and name_id = $2;',[id, name_id],[:int4, :int4])
		else
			Backend.sql('select id from xenodot.node where (root_id, xenodot.parent_hierarchy(hierarchy)) = (select root_id, hierarchy from xenodot.node($1));',[id],[:int4])
		end
		if res
			nodes = res.collect{ |id| Node[id[0]]}
			if nodes.size == 1
				return nodes[0]
			else
				return nodes
			end
		end
	end
	
	def fast_replace(*args)
		name_params, options, value = NodeInstanceMethods.parse_qname_and_options(args, :has_value => true)
        name_id = QName.id(*name_params)
       	Backend.sql("update xenodot.property set value = #{value.sql_value} where node_id = #{self.id} and name_id = #{name_id}",[],[])
	end

    
  end
  
  class XPathBuilder
    include NodeInstanceMethods
    attr_reader :start_node_id
    
    def xpath
      if @condition
        "#{@xpath}[#{@condition}]"
      else
        @xpath
      end
    end
    
    def initialize(parent, fragment, condition = nil)
      case parent
      when Integer:
        @start_node_id = parent
        @xpath = fragment
        @condition = condition
      when Node:
        @start_node_id = parent.id
        @xpath = fragment
        @condition = condition
      when XPathBuilder:
        @start_node_id = parent.start_node_id
        @xpath = "#{parent.xpath}/#{fragment}"
        @condition = condition
      else
        raise "Can not handle parent #{parent.inspect}"
      end
    end
    
    def where(condition)
      return XPathBuilder.new(start_node_id, @xpath, condition)
    end
    
    def exist?
      size > 0
    end
    
    def size
      Backend.sql('select count(*) from xpath.node(xenodot.node($1), $2)', [start_node_id, xpath], [:int4, :text], :count => 1)[0]
    end
    
    def delete!(options = {:array => false})
      deleted = Backend.sql(
        'delete from xenodot.node where id in (select id from xpath.node(xenodot.node($1), $2)) returning id',
        [@start_node_id, @xpath], [:int4, :text]
      )
      return options[:array] ? [] : nil unless deleted
      deleted = deleted.collect{|row| Node.new(row[0])}
      deleted = deleted[0] unless options[:array] or deleted.size > 1
      return deleted
    end
        
    def id(options = {:array => false})
      result = Backend.sql('select id from xpath.node(xenodot.node($1), $2)', [start_node_id, xpath], [:int4, :text])
      if result
        result = result.collect{|row| row[0]}
        result = result[0] unless result.size > 1 or options[:array]
      end
      result
    end
    
    def materialize!(options = {:array => false})
      result = id(:array => true).collect{|node_id| Node.new(node_id)}
      result = result[0] unless result.size > 1 or options[:array]
      result
    end
    
    def sql_value
    	"xenodot.value_path('#{@start_node_id.to_s + '|' + self.xpath}')"
    end
    
    def == other
    	@start_node_id == other.start_node_id and self.xpath == other.xpath
    end
  end
end