module Xenodot
  module Query
    module JcrPath
      def self.node_query(start_node, path)
        query = AbstractQuery.new(nil) do |sql|
          sql.from            = 'xenodot.node as current_node'
          if start_node
            sql.where         = 'current_node.id = ?'
            sql.where_params  = start_node['id']
            sql.where_types   = :int4
          end
        end
        path = path.strip.split('/')
        if path[0] == ''
          query = AbstractQuery.new(query) do |sql|
            sql.from          = 'xenodot.node as current_node'
            sql.where         = %q{current_node.root_id = parent1_node.root_id and current_node.hierarchy = ''}
          end
          path.delete_at(0)
        end
        path.each_with_index do |token, ii|
          case token
          when '.':
            #ignore
          when '..':
            query = AbstractQuery.new(query) do |sql|
              sql.from        = 'xenodot.node as current_node'
              sql.where       = 'current_node.root_id = parent1_node.root_id and xenodot.parent_hierarchy(parent1_node.hierarchy) = current_node.hierarchy'
            end
          else
            prefix, local_name, offset = parse_name(token)
            
            query = AbstractQuery.new(query) do |sql|
              sql.from        = 'xenodot.qname(?, ?) as current_node_name, xenodot.node as current_node'
              sql.from_params = [prefix, local_name]
              sql.from_types  = [:text, :text]
              sql.where       = %q{current_node.id = (select id from xenodot.node
                where     node.root_id = parent1_node.root_id
                      and node.name_id = current_node_name.id
                      and xenodot.parent_hierarchy(node.hierarchy) = parent1_node.hierarchy
                offset ?
                limit 1
              )}
              sql.where_params = [offset]
              sql.where_types  = [:int4]
            end
          end
        end
        query
      end
      
      def self.parse_name(token)
        elements = token.match(/((\w+):)?(\w+)(\[(\d+)\])?\Z/)
        raise "Syntax error, #{token.inspect} does not match" unless elements
        prefix = elements[2] || ''
        local_name = elements[3]

        if elements[5]
          offset = elements[5].to_i
          raise "Syntax error, index must be > 0" if offset < 1
          offset -= 1
        else
          offset = -1
        end
        return prefix, local_name, offset
      end
      
    end
  end
end