module Xenodot
  
  class MinimizedSlate
    instance_methods.each { |m| undef_method m unless m =~ /^__|^respond_to?|^inspect|^to_s|^kind_of?|^class|^should|^instance_of?|^kind_of?|object_id|equals?|^send|^nil|^method|^_method_missing/ }
  end

  module Query
    class SQLParams
      attr_accessor :from, :where, :from_params, :where_params, :from_types, :where_types, :default, :default_type

      def initialize
        yield self
        
        [:from_params, :where_params, :from_types, :where_types].each do |focus|
          curr_val = self.send(focus)
          self.send "#{focus}=", case curr_val
            when nil: []
            when Array: curr_val
            else [curr_val]
          end
        end
        
        self.default      ||= :'current_node.*'
        self.default_type ||= :'xenodot.node'
      end
    end
    
    module SQLGenerator
      def sql!(options = {}, &block)
        stack = []
        stack << SQLParams.new(&block) unless block.nil?

        current = self
        begin
          stack << current.sql?
        end while (current.respond_to?(:parent_query!) and current = current.parent_query! and current.respond_to?(:sql?))

        prefix = options.delete(:prefix)
        
        param_count = 0        
        config = Hash.new{|hash, key| hash[key] = []}
        [:from, :where].each do |accessor|
          stack.each_with_index do |row, index|
            curr_val = row.__send__(accessor)
            if curr_val
              curr_val = curr_val.gsub(/parent(\d+)_/){ |key| "#{prefix}parent#{$1.to_i + index}_" }
              if index > 0
                curr_val = curr_val.gsub(/current_/, "#{prefix}parent#{index}_")
              elsif prefix
                curr_val = curr_val.gsub(/current_/, "#{prefix}current_")
              end
              curr_val = curr_val.gsub('TRUEROOT', 'current_node')
              curr_val = curr_val.gsub('TRUEPARENT', 'parent1_node')
              unless options[:no_exec]
                config[accessor] << curr_val.gsub(/\?/){|param| "$#{param_count += 1}"}
              end
              config["#{accessor}s".to_sym] << curr_val unless curr_val == nil or curr_val.empty?
            end
            [:params, :types].each do |helper|
              key = "#{accessor}_#{helper}".to_sym
              config[key] += row.send(key)
            end
          end
        end
        config[:default] = stack[0].default
        config[:default_type] = stack[0].default_type
        
        [:from_params, :where_params, :from_types, :where_types].each do |focus|
          config[focus].flatten!
          config[focus].compact!
        end
        
        if options.delete(:no_exec)
          config[:default] = config[:default].to_s.gsub(/current_/, "#{prefix}current_") if prefix
          if options.delete(:value) and not (config[:default_type] == :literal or config[:default_type] == :primitive)
            config[:default] = "xpath.string(#{config[:default]})"
          end
          return config
        else
          sql = ''
          if config[:from].size > 0
            sql << "from #{config[:from].join(', ')}"
          end
          if config[:where].size > 0
            sql << ' where ' << config[:where].join(' and ')
          end
          sql << " order by #{options[:order_by]}" if options[:order_by]
          return Backend.sql("select #{config[:default]} #{sql}", config[:from_params] + config[:where_params], config[:from_types] + config[:where_types], options)
        end
      end
    end
    
    class AbstractQuery < MinimizedSlate
      include SQLGenerator
      
      def initialize(parent, &block)
        @parent = parent
        @sqlparams = SQLParams.new(&block) unless block.nil?
      end

      def parent_query!
        @parent
      end

      def sql?
        @sqlparams
      end
    end
  end
end