xml = %q{
<ancesto>
	<p1>
	<precedin/>
	</p1>
	<paren>
	<precsibling/>
	<sel>
		<child1>
		<desc1/>
		<desc2/>
		</child1>
		<child2>
		<desc3/>
		</child2>
	</sel>
	<folsibling/>
	</paren>
	<p2>
	<followin/>
	</p2>
</ancesto>
}.gsub(/\s/, '')


describe Xenodot::Node do

  before(:each) do
    reset_xenodot_database
    @node = Xenodot::Node['http://xenodot.org/test']
    @node.append_xml(xml)
    @self = Xenodot::Node[Xenodot::Backend.sql("select node_id from xenodot.dom where node_local_name = $1;",['sel'],[:text], :count => 1 )[0]]
  end
  
  it "should know its parent" do
  	@self.parent.size.should == 1
  	@self.parent.name.local_name.should == 'paren'
  end
    
  it "should know its children" do
  	@self.child.size.should == 2
  	@self.child[0].name.local_name.should == 'child1'
  	@self.child[1].name.local_name.should == 'child2'
  end

  it "should know its descendants" do
  	@self.descendant.size.should == 5
  	@self.descendant[0].name.local_name.should == 'child1'
  	@self.descendant[1].name.local_name.should == 'desc1'
  	@self.descendant[2].name.local_name.should == 'desc2'
  	@self.descendant[3].name.local_name.should == 'child2'
  	@self.descendant[4].name.local_name.should == 'desc3'
  end

  it "should know descendants or self" do
  	@self.descendant_or_self.size.should == 6
  	@self.descendant_or_self[0].name.local_name.should == 'sel'
  	@self.descendant_or_self[1].name.local_name.should == 'child1'
  	@self.descendant_or_self[2].name.local_name.should == 'desc1'
  	@self.descendant_or_self[3].name.local_name.should == 'desc2'
  	@self.descendant_or_self[4].name.local_name.should == 'child2'
  	@self.descendant_or_self[5].name.local_name.should == 'desc3'
  end

  it "should know its ancestor" do
  	@self.ancestor.size.should == 3
  	@self.ancestor[0].name.local_name.should == '!document-root!'
  	@self.ancestor[1].name.local_name.should == 'ancesto'
  	@self.ancestor[2].name.local_name.should == 'paren'
  end

  it "should know ancestor or self" do
  	@self.ancestor_or_self.size.should == 4
  	@self.ancestor_or_self[0].name.local_name.should == '!document-root!'
  	@self.ancestor_or_self[1].name.local_name.should == 'ancesto'
  	@self.ancestor_or_self[2].name.local_name.should == 'paren'
  	@self.ancestor_or_self[3].name.local_name.should == 'sel'
  end

  it "should know all preceding nodes" do
  	@self.preceding.size.should == 3
  	@self.preceding[0].name.local_name.should == 'p1'
  	@self.preceding[1].name.local_name.should == 'precedin'
  	@self.preceding[2].name.local_name.should == 'precsibling'
  end

  it "should know all preceding siblings" do
  	@self.preceding_sibling.size.should == 1
  	@self.preceding_sibling.name.local_name.should == 'precsibling'
  end

  it "should know all following nodes" do
  	@self.following.size.should == 3
  	@self.following[0].name.local_name.should == 'folsibling'
  	@self.following[1].name.local_name.should == 'p2'
  	@self.following[2].name.local_name.should == 'followin'
  end

  it "should know all following siblings" do
  	@self.following_sibling.size.should == 1
  	@self.following_sibling.name.local_name.should == 'folsibling'
  end

    
    
end

describe Xenodot::XPathBuilder do


  before(:each) do
    reset_xenodot_database
    @node = Xenodot::Node['http://xenodot.org/test']
    @node.append_xml(xml)
    @self = @node.descendant('sel')
  end
  
  it "should know its parent" do
  	@self.parent.size.should == 1
  	@self.parent.name.local_name.should == 'paren'
  end
    
  it "should know its children" do
  	@self.child.size.should == 2
  	@self.child[0].name.local_name.should == 'child1'
  	@self.child[1].name.local_name.should == 'child2'
  end

  it "should know its descendants" do
  	@self.descendant.size.should == 5
  	@self.descendant[0].name.local_name.should == 'child1'
  	@self.descendant[1].name.local_name.should == 'desc1'
  	@self.descendant[2].name.local_name.should == 'desc2'
  	@self.descendant[3].name.local_name.should == 'child2'
  	@self.descendant[4].name.local_name.should == 'desc3'
  end

  it "should know descendants or self" do
  	@self.descendant_or_self.size.should == 6
  	@self.descendant_or_self[0].name.local_name.should == 'sel'
  	@self.descendant_or_self[1].name.local_name.should == 'child1'
  	@self.descendant_or_self[2].name.local_name.should == 'desc1'
  	@self.descendant_or_self[3].name.local_name.should == 'desc2'
  	@self.descendant_or_self[4].name.local_name.should == 'child2'
  	@self.descendant_or_self[5].name.local_name.should == 'desc3'
  end

  it "should know its ancestor" do
  	@self.ancestor.size.should == 3
  	@self.ancestor[0].name.local_name.should == '!document-root!'
  	@self.ancestor[1].name.local_name.should == 'ancesto'
  	@self.ancestor[2].name.local_name.should == 'paren'
  end

  it "should know ancestor or self" do
  	@self.ancestor_or_self.size.should == 4
  	@self.ancestor_or_self[0].name.local_name.should == '!document-root!'
  	@self.ancestor_or_self[1].name.local_name.should == 'ancesto'
  	@self.ancestor_or_self[2].name.local_name.should == 'paren'
  	@self.ancestor_or_self[3].name.local_name.should == 'sel'
  end

  it "should know all preceding nodes" do
  	@self.preceding.size.should == 3
  	@self.preceding[0].name.local_name.should == 'p1'
  	@self.preceding[1].name.local_name.should == 'precedin'
  	@self.preceding[2].name.local_name.should == 'precsibling'
  end

  it "should know all preceding siblings" do
  	@self.preceding_sibling.size.should == 1
  	@self.preceding_sibling.name.local_name.should == 'precsibling'
  end

  it "should know all following nodes" do
  	@self.following.size.should == 3
  	@self.following[0].name.local_name.should == 'folsibling'
  	@self.following[1].name.local_name.should == 'p2'
  	@self.following[2].name.local_name.should == 'followin'
  end

  it "should know all following siblings" do
  	@self.following_sibling.size.should == 1
  	@self.following_sibling.name.local_name.should == 'folsibling'
  end

    
    
end

