#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

module PiVM
module Pi

	# TODO implement support for key value pairs for communication
	def self.write(process)
		activity = process.current_activity
		process.current_activity = activity.next
		process.substitute(activity[:names])
	end

	def self.read(process, *args)
		args = args[0] if args.size == 1 and args[0].kind_of? Array
		activity = process.current_activity
		names = activity[:names]
		s = (names.kind_of? Array) ? names.size : 1
		#p activity.to_s
		raise("Wrong number of arguments, expected #{s} but got #{args.size}. Expected params are #{names.join(', ')}, got #{args.join(', ')} #{args.inspect}") unless s == args.size
		process.substitute!(names,(s == 1) ? args[0] : args)
		process.current_activity = activity.next
	end

	def self.fresh(process)
		activity = process.current_activity
		names = activity[:names]
		names = [names] unless names.kind_of? Array
		names.each do |name|
			n = PiVM.fresh_name(name)
			process.substitute!(name, n)
		end
		process.current_activity = activity.next
		process
	end

	def self.inaction(process)
		process.delete!
		false
	end


	def self.replication(process)
		activity = process.current_activity.next
		op = activity.type
		ch = process.substitute(activity[:channel])
		Queue.enqueue(process, op, ch, -1, true)
		while p2 = Queue.dequeue(activity.complement, ch, false) do
			p1 = Pi.replication2(process)
			PiVM.communicate(p1,p2)
			Queue.enqueue(p1)
			Queue.enqueue(p2)
		end
	end

	def self.replication2(process)
		# copy the process
		p = process.clone
		# remove the replication node
		p.current_activity = p.current_activity.next
		p
	end
	
	def self.tau(process)
		activity = process.current_activity
		function = activity[:function]
		if function
			inn = process.substitute(activity[:in])
			p "VM: Call function #{function} with parameters #{inn.inspect}" if $DEBUG
			res = if inn
					Functions.send(function, inn)
				else
					Functions.send(function)
				end
			out = activity[:out]
			ressize = (res.kind_of? Array) ? res.size : (res.nil?) ? 0 : 1
			outsize = (out.kind_of? Array) ? out.size : (out.nil?) ? 0 : 1
			raise "Wrong number of return values" unless ressize >= outsize 
			process.substitute!(out, res) if outsize > 0
			p "VM: Function #{function} retruned with #{out.inspect}" if $DEBUG
		end
		process.current_activity = activity.next
		process
	end
	
	# process continue as the first parallel process
	# for all other processes a new process-node is created 
	# and the processes will be enqueued
	def self.par(process)
		activity = process.current_activity
		childs = activity.next
		process.current_activity = childs[0]
		childs[1..-1].each do |child|
			p = process.clone
			p.current_activity = child
			Queue.enqueue(p)
		end
		process
	end
	
	# try to find a co
	def self.sum(process)
		activity = process.current_activity
		childs = activity.next
		# remove branches with negative matches and search for taus
		possible_processes = childs.collect do |act|
			act.act_as Activity # TODO
			# check all match operators, more then one match operator can be chained
			res = true
			while (op = act.type) == "match" do
				if Pi.match(process,act)
					act = act.next
				else
					# this subtree is no choice for the summation
					res = nil
					break
				end
			end
			if res
				if op == 'tau'
					# found a tau, decide for this branch
					process.current_activity = act
					return process
				else
					
					
					ch = process.substitute(act[:channel])
					if p2 = Queue.dequeue(act.complement, ch)
						# found an enabled communication, choose this branch
						process.current_activity = act
						PiVM.communicate(process, p2)
						Queue.enqueue(p2)
						return process
					end
					act
				end
			end
		end.compact
		
		# possible_processes contains only reads and writes with
		# no communication partner, so enque all of them
		possible_processes.each do |act|
			ch = process.substitute(act[:channel])
			Queue.enqueue(process, act.type, ch)
		end
		false
	end

	def self.call(process)
		activity = process.current_activity
		# Create a new process
		agent = PiVM.agent(activity[:agent])
		# Set the parameters right
		act_params = activity[:params]
		agent_params = agent[:params]
		
		raise "wrong number of parameters, expected #{agent_params.size} but got #{act_params.size}!" unless (agent_params.nil? and act_params.nil?) or (agent_params.size == act_params.size) #or (!agent_params.kind_of? Array and !act_params.kind_of? Array)
		
		unless agent_params.nil?
			process.substitute!(agent_params,process.substitute(act_params))
		end
		p "VM: calling Agent #{activity[:agent]}" if $DEBUG
		
		process.current_activity = agent.start
# 		
# 		res = Xenodot::Backend.sql(%q{select child.id from xenodot.node($1) as parent, xenodot.node as child, xenodot.property where 
# 			child.root_id = parent.root_id 
# 			and xenodot.parent_hierarchy(child.hierarchy) = parent.hierarchy
# 			and property.node_id = child.id
#       		and (property.value).content = $2;},[PiVM.agents.id, activity[:agent]],[:int4,:bytea], :count => 1)
# 		#res = Xenodot::Backend.sql(%q{select id from xenodot.node where hierarchy like (select node_hierarchy || '_' from xenodot.dom as d where d.node_hierarchy like (select hierarchy || '_'  from xenodot.node( $1 )) and d.value_content = $2)},[PiVM.agents.id, node[:agent]],[:int4,:bytea], :count => 1)
# 		process.current_activity = Xenodot::Node[res[0]].fast_child
# 		#a = PiVM.agents.child.where(%Q{@name = "#{node[:agent]}"}).materialize!
# 		#process[:pi] = a.fast_child
		process
	end
	
	def self.match(process, activity = nil)
		test_only = !activity.nil?
		activity ||= process.current_activity
		
		x = process.substitute(activity[:x])
		y = process.substitute(activity[:y])
		warn "Does #{x} and #{y} match? #{(x == y)? "Yes!" : "No!"}" if $DEBUG
		return x == y if test_only
		
		if x == y
			process.current_activity = activity.next
			process
		else
			Pi.inaction(process)
		end
	end

end
end
