#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#require 'Lexer'

class String
	
	def each_chr_with_index
		i = 0
		self.each_byte do |b|
			b = b.chr
			yield(b,i)
			i = i.next
		end
	end

end

module PiVM
 module PiParser
 
 	NAME = "[[:lower:]][[:alnum:]_$'\/]*"
 	AGENT = "[[:upper:]][[:alnum:]_$']*"
 	
 	def self.create_pi_parser
 		@pi_lexer = Lexer.new([
 			[/\Aagent.*?=/, ?d], #agent definition
 			[/\At\b(\{.*?\})?(\<.*?\>)?(\(.*?\))?/, ?t], #tau
			[/\A#{NAME}(\(.*?\))?/, ?n], #name
			[/\A(´|’|')#{NAME}(<.*?>)?/, ?c], #coname
			[/\A#{AGENT}(\(.*?\))?/, ?a], #agent call
			[/\A\|/, ?|],
			[/\A\+/, ?+],
			[/\A\((\^|ˆ).*?\)/, ?v],
			[/\A\./, ?.],
			[/\A\(/, ?(],
			[/\A\)/, ?)],
			[/\A!/, ?!],
			[/\A0/, ?0],
			[/\A\[.*?\]/, ?m],
			])

 		@part_lexer = Lexer.new([
 			[/\Aagent/, ?a],
 			[/\A#{AGENT}/, ?A], # agent name
 			[/\A\(/, ?(],
 			[/\A#{NAME}/, ?n], #param
 			[/\A,/, ?,],
 			[/\A\)/, ?)],
 			[/\A=/, ?=],
 			[/\A(´|’|')/, ?c],
 			[/\A</, ?<],
 			[/\A\>/, ?>],
 			[/\A(\^|ˆ)/,?v]])
 			
 		@tau_lexer = Lexer.new([
 			[/\A#{NAME}/, ?n], #param
 			[/\A,/, ?,],
 			[/\A\(/, ?(],
 			[/\A\)/, ?)],
 			[/\A\{/, ?{],
 			[/\A\}/, ?}],
 			[/\A</, ?<],
 			[/\A\>/, ?>],
 			])

 			
 	end
 	
 	%w{pi def name}.each do |n|
 		eval  %Q{
 	def self.#{n}_lexer()
 		@#{n}_lexer
 	end
 	}
 	end
 	
 	def self.parse_file(file)
 		root = PiVM.sandbox.create_child(:pi_parser)
 		File.open(file).each do |line|
 			PiParser.parse_pi(line,root)
 		end
 		root
 	end
 	
 	
 	def self.parse_pi( s , root = nil)
 		PiParser.create_pi_parser unless @pi_lexer
 		
 		s = s.gsub(/\s/,"").strip
 		p s
 		lex, tokens = @pi_lexer.scan(s)
 		lex << ')'
 		root = PiVM.sandbox.create_child(:pi_parser) unless root
 		agent = nil
 		current = nil
 		scopes = []
		lex.each_chr_with_index do |c,i|
			#p "#{c} #{i} c:#{current} s:#{scopes}"
 			case c
 			when 'd'
 				agent = PiParser.add_agent(tokens[i],root,"agent")
 				current = root.create_child(:tmp)
 				scopes << { :node => current, :parent => agent}
 			when '(':
 				parent = current
 				current = root.create_child(:tmp) 
 				scopes << { :node => current, :parent => parent }
 			when 'n': current = PiParser.add_read_write(tokens[i],current,"read")
 			when 'c': current = PiParser.add_read_write(tokens[i],current,"write")
 			when 'a': current = PiParser.add_call(tokens[i],current,"call")
 			when 'v': current = PiParser.add_restriction(tokens[i],current)
 			when '0': current = current.create_child("inaction")
 			when 't': current = PiParser.add_tau(tokens[i],current)
 			when '!': current = current.create_child("replication")
 			when 'm': current = PiParser.add_match(tokens[i],current)
 			when '|':
 				scopes.last[:type] = 'par'
 				current = scopes.last[:node]
 			when '+':
 				scopes.last[:type] = 'sum'
 				current = scopes.last[:node]
 			when ')'
 				scope = scopes.slice!(-1)
				if scope[:type]
					tmp = scope[:parent].create_child(scope[:type])
				else
					tmp = scope[:parent]
				end
				scope[:node].child.each do |n|
					n.parent = tmp
				end
				scope[:node].delete!
				current = scopes.last[:node] unless scopes.empty?
				
 			end
 			
 		end
 		root
 	end
 	
 	def self.add_agent(s,parent,op)
		lex, tokens = @part_lexer.scan(s)
		node = parent.create_child(op)
		names = []
		lex.each_chr_with_index do |c,i|
			case c
			when 'A': node[:name] = tokens[i]
			when 'n': names << tokens[i]
			end
		end
		if names.size == 1 then node[:params] = names[0] end
		if names.size > 1 then node[:params] = names end
		node
 	end
 	
 	def self.add_call(s,parent,op)
		lex, tokens = @part_lexer.scan(s)
		node = parent.create_child(op)
		names = []
		lex.each_chr_with_index do |c,i|
			case c
			when 'A': node[:agent] = tokens[i]
			when 'n': names << tokens[i]
			end
		end
		if names.size == 1 then node[:params] = names[0] end
		if names.size > 1 then node[:params] = names end
		node
 	end
 	
 	
 	# parse read or write actions. The first parsed name is the
 	# subject, all further names are objects
	def self.add_read_write(s, parent,operation)
		lex, tokens = @part_lexer.scan(s)
		node = nil
		names = []
		lex.each_chr_with_index do |c,i|
			if c == 'n'
				unless node
					node = parent.create_child(operation)
					node[:channel] = tokens[i]
				else
					names << tokens[i]
				end
			end
		end
		if names.size == 1 then node[:names] = names[0] end
		if names.size > 1 then node[:names] = names end
		node
	end
	

	def self.add_restriction(s, parent)
		lex, tokens = @part_lexer.scan(s)
		node = parent.create_child("fresh")
		names = []
		lex.each_chr_with_index do |c,i|
			names << tokens[i] if c == 'n'
		end
		if names.size == 1 then node[:names] = names[0] end
		if names.size > 1 then node[:names] = names end
		node
	end
	
	def self.add_match(s, parent)
		a,b = s.gsub(/[\s\[\]]/,'').split('=')
		node = parent.create_child("match")
		node[:x] = a
		node[:y] = b
		node
	end
	
	def self.add_tau(s,parent)
		lex, tokens = @tau_lexer.scan(s)
		node = parent.create_child("tau")
		funcname = []
		innames = []
		outnames = []
		names = [] # to parse the t in
		lex.each_chr_with_index do |c,i|
			case c
				when '{': names = funcname
				when '<': names = innames
				when '(': names = outnames
				when 'n': names << tokens[i] 
			end
		end
		node[:function] = funcname[0] if funcname[0]
		if innames.size == 1 then node[:in] = innames[0] end
		if innames.size > 1 then node[:in] = innames end
		if outnames.size == 1 then node[:out] = outnames[0] end
		if outnames.size > 1 then node[:out] = outnames end
		node
	end
 end
end
