#
#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
 
module PiVM
  
	TYPES = { :process_id => :int4, :thread_id => :int4, :operation => :text, :channel => :text }
	
	 
# 	def self.init
# 		@root = Node['http://xenodot.org/vm']
# 		@processes = @root.ensure_child(:vm, :processes)
# 		@agents = @root.ensure_child(:vm, :agents)
# 		@intern_agents = @root.ensure_child(:vm, :intern_agents)
# 		@resp = @intern_agents.ensure_child(:vm, :resp)
# 		unless @resp.child('read').exist?
# 			@resp.append_xml('<read channel="backchannel" names="resp"><inaction /></read>')
# 		end
# 		@sandbox = @root.ensure_child(:vm, :sandbox)
# 		@name = QName['names']
# 	end
# 	
# 	%w{agents processes root sandbox name}.each do |n|
# 		eval  %Q{
# 			def self.#{n}
# 				PiVM.init unless @#{n}
# 				@#{n}
# 			end
# 		}
# 	end

	# Load a pi-process from an xml file into the DOM
	def self.load(file)
		autostart = []
		raise "file does not exist" unless File.exist?(file)
		case File.extname(file)
		when ".xml"
			# load the file into a sandbox and remove all whitespace text nodes
			node = @sandbox.append_xml(File.read(file).gsub(/>\W+</,'><'))
		when ".pi"
			node = PiVM::PiParser.parse_file(file)
		else
			raise "Unknown file extension: #{File.extname(file)}"
		end
		#warn node.to_xml
		#p node.descendant('agent').size
		node.descendant(:agent).each do |agent|
			#puts agent.to_xml
			# do some checks on the process(es) (is the process valid)
			raise "An agent definition must provide a name for the agent" if agent[:name].nil? or agent[:name].empty?
			raise "An agent definition must have one child" unless agent.child.size == 1
			agent.descendant_or_self.each do |n|
				n.act_as Activity
				PiVM.parse_attributes(n)
				PiVM.check_consistency(n)
			end

			
			# move the process to the correct position in the DOM
			agent.parent = @agents
			
			autostart << process if agent[:autostart]
		end
		
		#clean up sandbox
		node.delete!

		# start porcesses marked with autostart
		autostart.each do |process|
			process = PiVM.create_process(agent.next)
			PiVM.handle_process(process)
		end

	end
	
	def self.check_consistency(activity)
		#warn activity.to_xml
		case activity.type
		when "inaction"
			raise "Inaction must not have a childs" unless activity.next.size == 0
		when "call"
			raise "Agent calls must not have a childs" unless activity.next.size == 0
			raise "Agent calls must name an agent to call" if activity[:agent].nil? or activity[:agent].empty?
		when "match"
			raise "A match should have two names to compare" if activity[:x].nil? or activity[:x].empty? or activity[:y].nil? or activity[:y].empty?
			raise "A match must have one child" unless activity.next.size == 1
			raise "A match must be followed by a match, read, write or tau" unless ["match","read","write","tau"].include? activity.next.type
		when "par"
			raise "A parallel split must have at least two childs" unless activity.next.size >= 2
		when "sum"
			raise "A summation must have at least two childs" unless activity.next.size >= 2
		when "read"
			raise "A read must have one child" unless activity.next.size == 1
		when "write"
			raise "A write must have one child" unless activity.next.size == 1
		when "replication"
			raise "A replication must have one child" unless activity.next.size == 1
		when "tau"
			raise "A tau must have one child" unless activity.next.size == 1
		when "fresh"
			raise "A fresh must have one child" unless activity.next.size == 1
		when "agent"
			raise "An agent must have one child" unless activity.next.size == 1
		else
			raise "unknown operation #{activity.type}"
		end
	end
	
	def self.parse_attributes(node)
		node.attributes.each do |key|
			# TODO correct the following line
			key = key.local_name
			value = node[key]
			value = value.split(",") unless value.kind_of? Array
			value = value.collect do |v|
				if v.starts_with? 'http://'
					URI.parse(v)
				elsif v.starts_with? 'http/'
					v[5..-1]
				else
					v
				end
			end
			value = value[0] if value.size == 1
			node[key] = value
		end
	end
	
	# handle one step of a process
	# return ture if process has another step to handle, else false
	def self.handle_process(process)
		begin
		activity = process.current_activity
		op = activity.type
		warn "Process #{process.id} activity #{op}" if $DEBUG
		if op == 'read' or op == 'write'
			channel = activity[:channel]
			# look for name conversions
			channel = process.substitute(channel)
			if op == 'write' and channel.kind_of? URI and PiVM.is_extern?(channel)
				PiVM.communicate_extern(process, channel)
			else
				op2 = activity.complement
				p2 = Queue.dequeue(op2,channel)
				if p2
					#handle replications while expecting a communication
					activity2 = p2.current_activity
					warn "activity2 is nil" if activity2.nil?
					op2 = activity2.type
					if op2 == 'replication'
						p2 = Pi.replication2(p2)
					end
					PiVM.communicate(process, p2)
					# enque the other process so that it can be handled from other threads
					Queue.enqueue(p2)
					return process
				else
					Queue.enqueue(process, op, channel)
					return false
				end
			end
		else
			# handle all other operations
			warn op if $DEMO
			Pi.send(op,process)
		end
		rescue
			$!.message << "\n\tfrom " << $@.join("\n\tfrom ")
			$!.message << "\m\t" << activity.to_s if $DEBUG
			raise $!
		end
	end
	
	def self.is_extern?(uri)
		(uri.host =~ /xenodot.org$|localhost$/).nil?
	end

	def self.create_process(activity, *names)
		process = Process.new
		process.current_activity = activity
		process
	end

	
	# Do the communication between two processes
	def self.communicate(p1, p2)
		if p1.current_activity.type == 'read'
			p_read = p1
			p_write = p2
			warn "read\nwrite" if $DEMO
		else
			p_read = p2
			p_write = p1
			warn "write\nread" if $DEMO
		end
		args = Pi.write(p_write)
		Pi.read(p_read, args)
	end
	
	# do an extern communication. 
	# TODO: evaluate wether it is useful/possible to fork a new process
	def self.communicate_extern(process, uri)
		backchannel, verb, request = Pi.write(process)
		#Queue.enqueue(process)
		resp = Net::HTTP.start(uri.host, uri.port) {|http|
			http.send_request(verb, uri.path, request)
		}
		p2 = PiVM.create_process(@resp.child[0])
		p2.substitute!("backchannel",backchannel)
		p2.substitute!("resp",resp)
		PiVM.handle_process(p2)
		return process
	end

	# Generate unique names 
	def self.fresh_name(prefix)
		raise "Abstract method, has to be implemented"
	end
	

end
