#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# context "A PiVM::PiProcess" do
# 
#   specify "should give a complement to a operation" do
#     PiVM::PiProcess.complement(:read) == :write and PiVM::PiProcess.complement(:write) == :read
#   end
# end
include Xenodot


context "The database" do
	setup do
		reset_pivm_database
		PiVM.init
		Backend.sql('select pivm.vm_init()',[],[])
		PiVM.load(File.join(File.dirname(__FILE__),"pi_spec.xml"))
		Xenodot::Backend.sql('set client_min_messages to info;', [], [])
	end

	specify "should handle a process step on vm_next_step" do
		pi = PiVM.agent("P").start
		p = PiVM.create_process(pi)
		PiVM::Queue.enqueue(p)
		# handle fresh
		res = Backend.sql('select * from  pivm.vm_next_step($1,$2,$3)',[nil,123,'out'],[:int4,:int4,:text],:count => 1)
		res.should == [p.id, nil]
		# handle tau
		res = Backend.sql('select * from  pivm.vm_next_step($1,$2,$3)',[res[0],123,'out'],[:int4,:int4,:text],:count => 1)
		res.should == [p.id, nil]
		# handle write
		res = Backend.sql('select * from  pivm.vm_next_step($1,$2,$3)',[res[0],123,'out'],[:int4,:int4,:text],:count => 1)
		res.should == [nil, nil]
		# find output
		res = Backend.sql('select * from  pivm.vm_next_step($1,$2,$3)',[res[0],123,'out'],[:int4,:int4,:text],:count => 1)
		res.should == [nil, 'b']
		# handle inaction
		res = Backend.sql('select * from  pivm.vm_next_step($1,$2,$3)',[res[0],123,'out'],[:int4,:int4,:text],:count => 1)
		res.should == [nil, nil]
		Backend.sql("select * from xenodot.node where id = $1",[p.id],[:int4]).should be_false

		
	end
	
	teardown do
		Xenodot::Backend.sql('set client_min_messages to error;', [], [])
	end
end
# 
# module PiVM
# 	def self.new_connection
# 		PGconn.new('user' => 'xenodot','dbname' => 'xenodot_test')
# 	end
# 
# end
# 
# context "A PiVM" do
# 	setup do
# 		Xenodot::Backend.sql('delete from pivm.vm_queue', [], []);
# 		@conn2 = nil
# 	end
# 
# 	specify "should generate unique names on fresh_name" do
# 		prefix = "p"
# 		p1 = PiVM.fresh_name(prefix)
# 		p2 = PiVM.fresh_name(prefix)
# 		p1.should_not == p2
# 	end
# 
# 	specify "should be able to insert values in the wait queue" do
# 		Queue.enqueue(Node.new(123),'read','a',567)
# 		res = Xenodot::Backend.sql('select * from pivm.vm_queue',[],[])
# 		res[0][1..4].should == [ 123, 'read', 'a', 567 ]
# 	end
# 
# # TODO this test does not work without real multithreading
# # 	specify "should select the first row that is not locked on sql" do
# # 		@conn2 = PiVM.new_connection
# # 		Queue.enqueue(123,'read','a')
# # 		Queue.enqueue(124,'read','a')
# # 		@conn2.exec("start transaction")
# # # 		res1 = @conn2.query("select * from xenodot.vm_queue where channel = 'a' FOR UPDATE NOWAIT LIMIT 1")
# # 		res1 = @conn2.query("delete from xenodot.vm_queue where channel = 'a' and operation = 'read' returning *")
# # 		#res1 = PiVM.sql('select * from xenodot.vm_wait_queue where channel = $1 FOR UPDATE NOWAIT',['a'],[:text], :count => 1)
# # 		res1[0][1..3].should == [123, 'read', 'a']
# # 		res2 = Backend.sql('select * from xenodot.vm_dequeue($1,$2)',['read','a'],[:text,:text], :count => 1)
# # 		res2[1..3].should == [124, 'read', 'a']
# # 		
# # 		
# # 		
# # 	end
# 	
# 	specify "should be able to handle URIs as channel names" do
# 		pi = PiVM.agents.child.detect{|n| n[:name] == "URI"}.child
# 		p = PiVM.create_process(pi)
# 		PiVM.handle_process(p)
# 		p[:pi][:channel].should == URI.parse("http://localhost:8080")
# 		PiVM.handle_process(p)
# 		p2 = Queue.dequeue
# 		PiVM.handle_process(p2)
# 		p2.substitute("z").should == "b"
# 		
# 	end
# 	
# 	specify "should allow writes to extern services" do
# 		pi = PiVM.agents.child.detect{|n| n[:name] == "Extern"}.child
# 		p pi.to_xml
# 		p = PiVM.create_process(pi)
# 		PiVM.handle_process(p) # fresh
# 		PiVM.handle_process(p) # .should be_false #write
# 		PiVM.handle_process(p)
# 		puts Backend.sql("select * from pivm.vm_queue",[],[]).inspect
# 		#puts p.substitute("res").inspect
# 		p.substitute("res").should be_kind_of Net::HTTPResponse
# 	
# 	end
# 	
# 	teardown do
# 		if not @conn2.nil?
# 			@conn2.exec("commit")
# 			@conn2.close
# 		end
# 	end
# end
# 
# context "A PiVM::Process" do
# 
# 	setup do
# 		@node = PiVM.root.ensure_child('processtest')
# 		@names = @node.ensure_child('names')
# 		@names[:a] = 'b'
# 		@names[:x] = 'y'
# 		@node.act_as PiVM::Process
# 	end
# 	
# 	specify "should return then name on operation" do
# 		@node.operation.should == 'processtest'
# 	end
# 	
# 	specify "should return replaced name on map" do
# 		@node.substitute('a').should == 'b'
# 		@node.substitute('b').should == 'b'
# 		@node.substitute('c').should == 'c'
# 		@node.substitute('x').should == 'y'
# 		@node.substitute!('c','d')
# 		@node.substitute('c').should == 'd'
# 		@node.substitute!('c',nil)
# 		@node.substitute('c').should == 'c'
# 		@node.substitute(['a','b','c','x']).should == ['b','b','c','y']
# 		@node.substitute!(['a','b'],['z','o'])
# 		@node.substitute(['a','b','c','x']).should == ['z','o','c','y']
# 	end
# 	
# 	specify "should return the complement operation on self.complement" do
# 		PiVM::Process.complement('read').should == 'write'
# 		PiVM::Process.complement('wrtie').should == 'read'
# 	end
# 
# 	
# end
